package EFS::CLI::Any;

use strict;
use warnings;
use English;

use base q{EFS::CLI};
use base q{EFS::DBI};
use base q{EFS::DistableObject};

use Params::Validate qw( :all );

sub subitems {
    return shift->get_resultset( q{Metaproj} )->all;
}

sub history {

    my $class = shift;

    my (%args) = validate( @_, {
        %EFS::CLI::validate_common,
        %EFS::CLI::History::validation_specs,
    });

    return EFS::CLI::History->history(%args);

}

sub disttree {

    my $class = shift;

    my (%args) = validate( @_, {
        %EFS::CLI::validate_common,
        %EFS::CLI::validate_dist_args,
    });

    my @cells = $class->handle_cell_arguments(
        %args,
        include_parents => 1,
    );

    my @releases       = $class->get_resultset( q{Release} )->all;
    my @releasealiases = $class->get_resultset( q{ReleaseAlias} )->all;

    @releases = grep { not $_->destroyed } @releases;

    if ( grep { $_->celltype eq q{prod} } @cells ) {
        @releases       = grep { $_->has_only_prod_stages } @releases;
        @releasealiases = grep { $_->has_only_prod_stages } @releasealiases;
    }

    my $tree = EFS::DependencyTree->new({
        seeds => [ @releases, @releasealiases ],
        type  => q{runtime},
    });

    if ( $args{dryrun} ) {
        STDOUT->print(
            $tree->get_tree_text( cells => \@cells ),
        );
        return 1;
    } else {
        return $tree->dist( %args );
    }

}

sub purge {

    my $class = shift;

    my (%args) = validate( @_, {
        %EFS::CLI::validate_common,
        %EFS::CLI::validate_purge_args,
    });

    my @cells = $class->handle_cell_arguments(
        %args,
        default_scope   => q{global},
        include_parents => 1,
    );

    my @metaprojs = $class->get_resultset( q{Metaproj} )->all;

    STDOUT->print(qq{Purging all metaprojects\n});

    my $fm_metaprojs = EFS::ForkManager->new({

        parallel => 1,

        list => \@metaprojs,

        action => sub {

            my $metaproj = shift;

            my $name = $metaproj->name;

            return EFS::RPC->execute(
                class  => q{EFS::CLI::Metaproj},
                method => q{purge},
                args   => [
                    authorized => 1,
                    name       => $name,
                    %args,
                ],
                history => {
                    parentid => $args{history},
                    values   => {
                        request  => qq{purge metaproj $name},
                        metaproj => $name,
                    },
                },
            );

        },

    });

    $fm_metaprojs->run || die(
        qq{Unable to purge the following metaprojs:\n\t},
        join(qq{\n\t}, map { $_->name } $fm_metaprojs->get_failed ), qq{\n}
    );

    STDOUT->print( qq{Successfully purged all metaprojs\n} );

    return 1;

}

sub execlinks {

    my $class = shift;

    my (%args) = validate( @_, {
        %EFS::CLI::validate_common,
        %EFS::CLI::validate_diststate_args,
        force => { type => SCALAR | UNDEF, optional => 1 },
    });

    $class->assert_maintenance_mode;

    my @cells = $class->handle_cell_arguments(
        %args,
        default_scope => q{global},
    );

    my $fm_cells = EFS::ForkManager->new({

        list => \@cells,

        action => sub {

            my $cell = shift;

            return EFS::RPC->execute(
                cell   => $cell,
                class  => q{EFS::CLI::Cell},
                method => q{execlinks_rpc},
                args   => [
                    name => $cell->name,
                ],
            );

        },

    });

    $fm_cells->run || die(
        qq{Unable to update execlinks in the following cells:\n\t},
        join(qq{\n\t}, map { $_->name } $fm_cells->get_failed ), qq{\n}
    );

    STDOUT->print( qq{Updated execlinks in all cells\n} );

    return 1;

}

sub audit {

    my $class = shift;

    my (%args) = validate( @_, {
        %EFS::CLI::validate_common,
        audits  => { type     => SCALAR,
                     default  => q{required},
                     optional => 1 },
        force   => { type     => BOOLEAN,
                     optional => 1 },
        verbose => { type     => BOOLEAN,
                     optional => 1 },
    });

    my @metaprojs = $class->get_resultset( q{Metaproj} )->all;

    STDOUT->print(
        qq{Performing $args{audits} audits for all metaprojects\n}
    );

    my $fm_metaprojs = EFS::ForkManager->new({

        parallel => 4,

        list => \@metaprojs,

        action => sub {

            my $metaproj = shift;

            my $name = $metaproj->name;

            return EFS::RPC->execute(
                class  => q{EFS::CLI::Metaproj},
                method => q{audit},
                args   => [
                    authorized => 1,
                    name       => $name,
                    %args,
                ],
                history => {
                    parentid => $args{history},
                    values   => {
                        request  => qq{audit metaproj $name},
                        metaproj => $name,
                    },
                },
            );

        },

    });

    $fm_metaprojs->run || die(
        qq{Unable to sucessfully audit the following metaprojects:\n\t},
        join(qq{\n\t}, map { $_->name } $fm_metaprojs->get_failed ), qq{\n}
    );

    STDOUT->print( qq{Successfully audited all metaprojects\n} );

    return 1;

}

1;

# Local Variables:
#   mode: cperl
#   cperl-indent-level: 4
#   fill-column: 78
# End:
# vim: expandtab shiftwidth=4:
