package EFS::CLI::Fileqtree;

use strict;
use warnings;
use English;

use base q{EFS::CLI};
use base q{EFS::DBI};
use Params::Validate qw( :all );

sub init {

    my $class = shift;

    my (%args) = validate( @_, {
        %EFS::CLI::validate_common,
        cell       => { type     => SCALAR,
                        optional => 1 },
        fileserver => { type     => SCALAR,
                        depends  => [qw( cell )],
                        optional => 1 },
        fileshare  => { type     => SCALAR,
                        depends  => [qw( cell fileserver )],
                        optional => 1 },
        name       => { type     => SCALAR,
                        depends  => [qw( cell fileserver fileshare )],
                        optional => 1 },
    });

    my $resultset = $class->get_resultset;

    my @fileqtrees = ();

    ##  no critic: CascadingIfElse
    if ( not $args{cell} ) {

        @fileqtrees = $resultset->all;

    } elsif ( not $args{fileserver} ) {

        my $cell_rs = $class->get_resultset( q{Cell} );
        my $cell = $cell_rs->retrieve_bynames_assert( name => $args{cell} );
        $cell->assert_fstype( q{nfs} );
        @fileqtrees = $cell->fileqtrees;

    } elsif ( not $args{fileshare} ) {

        my $fileserver_rs = $class->get_resultset( q{Fileserver} );

        my $fileserver = $fileserver_rs->retrieve_bynames_assert(
            cell => $args{cell},
            name => $args{fileserver},
        );

        $fileserver->cell->assert_fstype( q{nfs} );

        @fileqtrees = $fileserver->fileqtrees;

    } elsif ( not $args{name} ) {

        my $fileshare_rs = $class->get_resultset( q{Fileshare} );

        my $fileshare = $fileshare_rs->retrieve_bynames_assert(
            cell       => $args{cell},
            fileserver => $args{fileserver},
            name       => $args{fileshare},
        );

        $fileshare->cell->assert_fstype( q{nfs} );

        @fileqtrees = $fileshare->fileqtrees;

    } else {

        $fileqtrees[0] = $resultset->retrieve_bynames_assert(
            cell       => $args{cell},
            fileserver => $args{fileserver},
            fileshare  => $args{fileshare},
            name       => $args{name},
        );

        if ( not $fileqtrees[0]->filetype ~~ [qw(dev dist)] ) {
            die qq{Not a dev or dist fileqtree\n};
        }

    }

    if ( not @fileqtrees ) {
        STDOUT->print( qq{No fileqtrees match the specified criteria\n} );
        return 1;
    }

    my $fm_fileqtrees = EFS::ForkManager->new({

        list => [ sort @fileqtrees ],

        action => sub {

            my $fileqtree = shift;

            return EFS::RPC->execute(
                cell   => $fileqtree->cell,
                class  => $class,
                method => q{init_rpc},
                args   => [
                    cell       => $fileqtree->cell->name,
                    fileserver => $fileqtree->fileserver->name,
                    fileshare  => $fileqtree->fileshare->name,
                    name       => $fileqtree->name,
                ],
            );

        },

    });

    $fm_fileqtrees->run || die(
        qq{Unable to initialize the following fileqtrees:\n\t},
        join(qq{\n\t}, $fm_fileqtrees->get_failed), qq{\n}
    );

    STDOUT->print( qq{Initialized all fileqtrees\n} );

    return 1;

}

sub init_rpc {

    my $class = shift;

    my (%args) = validate( @_, {
        %EFS::CLI::validate_common,
        cell       => { type       => SCALAR },
        fileserver => { type       => SCALAR },
        fileshare  => { type       => SCALAR },
        name       => { type       => SCALAR },
    });

    my $resultset = $class->get_resultset;

    my $fileqtree = $resultset->retrieve_bynames_assert(
        cell       => $args{cell},
        fileserver => $args{fileserver},
        fileshare  => $args{fileshare},
        name       => $args{name},
    );

    return $class->init_fileqtree( $fileqtree );

}

sub init_fileqtree {

    my $class = shift;

    my $fileqtree = shift;

    my $filetype = $fileqtree->filetype;

    my $admin_path = $fileqtree->admin_path;

    if ( not -d $admin_path ) {
        die(
            qq{No such directory $admin_path\n},
            qq{Make sure the autofs maps have are up to date.\n},
        );
    }

    if ( $filetype ne q{dist} ) {
        EFS::Util->MakeDir( path => qq{$admin_path/global} );
    }

    if ( not $filetype ~~ [qw( dev dist )] ) {
        STDOUT->print( qq{Initialized fileqtree $fileqtree\n} );
        return 1;
    }

    my $instancealias_rs = $class->get_resultset( q{PlatformInstanceAlias} );
    my $platformalias_rs = $class->get_resultset( q{PlatformAlias} );
    my %aliases = map { $_->name => $_ }
        ( $instancealias_rs->all, $platformalias_rs->all );

    foreach my $name ( sort keys %aliases ) {

        my $alias = $aliases{ $name };

        my $from = join q{/}, $admin_path, $name;
        my $to   = $alias->aliasof->name;

        if ( not -l $from and -d $from ) {
            EFS::Util->RenameDir(
                from => $from,
                to   => qq{$admin_path/$to},
            );
        }

        EFS::Util->MakeLink(
            from => $from,
            to   => $to,
        );

    }

    my $instance_rs = $class->get_resultset( q{PlatformInstance} );
    my %instances = map { $_->name => $_ } $instance_rs->all;

    my $platform_rs = $class->get_resultset( q{Platform} );
    my %platforms = map { $_->name => $_ } $platform_rs->all;

    my @subdirs = (
        $filetype,
        map { $_, qq{$_/$filetype} }
            ( sort keys %instances, sort keys %platforms ),
    );

    foreach my $subdir ( @subdirs ) {
        EFS::Util->MakeDir( path => qq{$admin_path/$subdir} );
    }

    my $io = IO::Dir->new( $admin_path )  ||
        die qq{Unable to opendir $admin_path: $ERRNO\n};

    foreach my $name ( sort $io->read ) {

        next if $name ~~ [ $filetype, qw(global . ..) ];

        # Real data lives in "type" directories, which have NO dots
        # Instance names ALWAYS have dots
        next if $name !~ m{\.}msx;

        my $path = qq{$admin_path/$name};

        if ( -l $path ) {
            if ( not $aliases{$name} ) {
                EFS::Util->RemoveLink( from => $path );
            }
        } elsif ( -d $path ) {
            if ( not $instances{$name} and not $platforms{$name} ) {
                EFS::Util->RemoveDirHierarchy( top => $path );
            }
        }

    }

    STDOUT->print(qq{Initialized fileqtree $fileqtree\n});

    return 1;

}

sub define {

    my $class = shift;

    my (%args) = validate( @_, {
        %EFS::CLI::validate_common,
        cell       => { type => SCALAR },
        fileserver => { type => SCALAR },
        fileshare  => { type => SCALAR },
        name       => { type => SCALAR },
    });

    my $fileshare_rs = $class->get_resultset( q{Fileshare} );

    my $fileshare = $fileshare_rs->retrieve_bynames_assert(
        cell       => $args{cell},
        fileserver => $args{fileserver},
        name       => $args{fileshare},
    );

    $fileshare->cell->assert_fstype( q{nfs} );

    my $resultset = $class->get_resultset;

    my ($fileqtree) = $resultset->search({
        fileshare => $fileshare->id,
        name      => $args{name},
    });

    if ( $fileqtree ) {
        STDOUT->print( qq{Definition of fileqtree $fileqtree exists\n} );
    } else {
        $fileqtree = $resultset->create({
            fileshare => $fileshare->id,
            name      => $args{name},
        });
        STDOUT->print( qq{Defined fileqtree $fileqtree\n} );
    }

    return $fileqtree;

}

sub undef {

    my $class = shift;

    my (%args) = validate( @_, {
        %EFS::CLI::validate_common,
        cell       => { type => SCALAR },
        fileserver => { type => SCALAR },
        fileshare  => { type => SCALAR },
        name       => { type => SCALAR },
    });

    my $resultset = $class->get_resultset;

    my $fileqtree = $resultset->retrieve_bynames_assert(
        cell       => $args{cell},
        fileserver => $args{fileserver},
        fileshare  => $args{fileshare},
        name       => $args{name},
    );

    my $object_id = $fileqtree->object_id;

    $fileqtree->attrs->delete_all;
    $fileqtree->delete;

    STDOUT->print( qq{Undefined qtree $object_id\n} );

    return 1;

}

sub display {

    my $class = shift;

    my (%args) = validate( @_, {
        %EFS::CLI::validate_common,
        cell       => { type     => SCALAR,
                        optional => 1 },
        fileserver => { type     => SCALAR,
                        depends  => [qw( cell )],
                        optional => 1 },
        fileshare  => { type     => SCALAR,
                        depends  => [qw( cell fileserver )],
                        optional => 1 },
        name       => { type     => SCALAR,
                        depends  => [qw( cell fileserver fileshare )],
                        optional => 1 },
    });

    my $resultset = $class->get_resultset;

    my @fileqtrees = ();

        ##  no critic: CascadingIfElse
    if ( not $args{cell} ) {

        @fileqtrees = $resultset->all;

    } elsif ( not $args{fileserver} ) {

        my $cell_rs = $class->get_resultset( q{Cell} );
        my $cell    = $cell_rs->retrieve_bynames_assert( name => $args{cell} );
        $cell->assert_fstype( q{nfs} );
        @fileqtrees = $cell->fileqtrees;

    } elsif ( not $args{fileshare} ) {

        my $fileserver_rs = $class->get_resultset( q{Fileserver} );

        my $fileserver = $fileserver_rs->retrieve_bynames_assert(
            cell => $args{cell},
            name => $args{fileserver},
        );

        $fileserver->cell->assert_fstype( q{nfs} );

        @fileqtrees = $fileserver->fileqtrees;

    } elsif ( not $args{name} ) {

        my $fileshare_rs = $class->get_resultset( q{Fileshare} );

        my $fileshare = $fileshare_rs->retrieve_bynames_assert(
            cell       => $args{cell},
            fileserver => $args{fileserver},
            name       => $args{fileshare},
        );

        $fileshare->cell->assert_fstype( q{nfs} );

        @fileqtrees = $fileshare->fileqtrees;

    } else {

        $fileqtrees[0] = $resultset->retrieve_bynames_assert(
            cell       => $args{cell},
            fileserver => $args{fileserver},
            fileshare  => $args{fileshare},
            name       => $args{name},
        );

    }

    $class->display_objects( objects => \@fileqtrees );

    if ( scalar @fileqtrees != 1 ) {
        return 1;
    }

    my $fileqtree = $fileqtrees[0];

    if ( my @attrs = $fileqtree->attrs ) {
        $attrs[0]->display_objects( objects => \@attrs );
    }

    return 1;

}

sub history {

    my $class = shift;

    my (%args) = validate( @_, {
        %EFS::CLI::validate_common,
        cell       => { type => SCALAR },
        fileserver => { type => SCALAR },
        fileshare  => { type => SCALAR },
        name       => { type => SCALAR },
        %EFS::CLI::History::validation_specs,
    });

    $args{fileqtree} = delete $args{name};

    return EFS::CLI::History->history(%args);

}

sub execlinks {

    my $class = shift;

    my (%args) = validate( @_, {
        %EFS::CLI::validate_common,
        cell       => { type => SCALAR },
        fileserver => { type => SCALAR },
        fileshare  => { type => SCALAR },
        name       => { type => SCALAR },
        force      => { type => SCALAR | UNDEF, optional => 1 },
    });

    $class->assert_maintenance_mode;

    my $resultset = $class->get_resultset;

    my $fileqtree = $resultset->retrieve_bynames_assert(
        cell       => $args{cell},
        fileserver => $args{fileserver},
        fileshare  => $args{fileshare},
        name       => $args{name},
    );

    return EFS::RPC->execute(
        cell   => $fileqtree->cell,
        class  => $class,
        method => q{execlinks_rpc},
        args   => [ %args ],
    );

}

sub execlinks_rpc {

    my $class = shift;
    my %args  = @_;

    my $resultset = $class->get_resultset;

    my $fileqtree = $resultset->retrieve_bynames_assert(
        cell       => $args{cell},
        fileserver => $args{fileserver},
        fileshare  => $args{fileshare},
        name       => $args{name},
    );

    $fileqtree->assert_exectype;

    $class->init_fileqtree( $fileqtree );

    my $platform_data = EFS::ExecLinks->get_all_platform_data;
    my $release_data  = EFS::ExecLinks->get_all_release_data;

    my $fm_execdirs = EFS::ForkManager->new({

        list => [ sort $fileqtree->get_execdirs ],

        action => sub {

            my $execdir = shift;

            return $class->execlinks_execdir(
                execdir       => $execdir,
                release_data  => $release_data,
                platform_data => $platform_data,
                force         => $args{force},
            );

        },
    });

    $fm_execdirs->run || die(
        qq{Unable to update the execlinks for the following:\n\t},
        join( qq{\n\t}, $fm_execdirs->get_failed ), qq{\n}
    );

    STDOUT->print( qq{Updated execlinks for $fileqtree\n} );

    return 1;

}

sub execlinks_execdir {

    my $class = shift;
    my %args  = @_;

    my $execdir       = $args{execdir};
    my $release_data  = $args{release_data};
    my $platform_data = $args{platform_data};

    my ($type,$admin_path,$cell,$release) =
        EFS::ExecLinks->parse_exec_dir( $execdir );

    my $disable_exec_backwards = EFS::ExecLinks->get_release_attr(
        data      => $release_data,
        release   => $release,
        attribute => q{disable_exec_backwards},
    );

    my $disable_exec_emulates = EFS::ExecLinks->get_release_attr(
        data      => $release_data,
        release   => $release,
        attribute => q{disable_exec_emulates},
    );

    my $locked = EFS::ExecLinks->get_release_attr(
        data      => $release_data,
        release   => $release,
        attribute => q{locked},
    );

    $locked = 0 if $args{force};

    EFS::ExecLinks->update_links(
        directory              => $execdir,
        platform_data          => $platform_data,
        disable_exec_backwards => $disable_exec_backwards,
        disable_exec_emulates  => $disable_exec_emulates,
        locked                 => $locked,
    );

    return 1;

}

1;

# Local Variables:
#   mode: cperl
#   cperl-indent-level: 4
#   fill-column: 78
# End:
# vim: expandtab shiftwidth=4:
