package EFS::CLI::Namespace;

use strict;
use warnings;
use English;

use feature q{switch};

use base q{EFS::CLI};
use base q{EFS::DBI};

use Params::Validate qw( :all );
use File::Basename qw(dirname);
use File::Glob q{:glob};

sub rebuild {

    my $class = shift;

    my (%args) = validate( @_, {
        %EFS::CLI::validate_common,
        type => { type => SCALAR },
        %EFS::CLI::validate_dist_args,
    });

    if ( $args{type} ne q{dev} && $args{type} ne q{home} ) {
        die qq{The only supported namespaces are 'dev' and 'home'\n};
    }

    my $project_rs = $class->get_resultset( q{Project} );

    my $project = $project_rs->retrieve_bynames_assert(
        metaproj => q{efs},
        name     => q{config-namespace},
    );

    return EFS::RPC->execute(
        cell   => $project->cell,
        class  => $class,
        method => q{rebuild_rpc},
        args   => [ %args ],
    );

}

sub rebuild_rpc {

    my $class = shift;
    my %args  = @_;

    my @cells =
        grep { not $_->celltype eq q{prod} }
            $class->handle_cell_arguments( %args, default_scope => q{global} );

    if ( not @cells ) {
        die qq{No dev cells match the specified criteria\n};
    }

    my $history_rs = $class->get_resultset( q{History} );
    my $history = $history_rs->retrieve_assert( id => $args{history} );

    my @moves = $history_rs->search({
        command => [ q{move project}, q{move home} ],
        status  => undef,
    });

    if ( @moves ) {

        STDOUT->print( qq{The following move commands are still active:\n} );

        $moves[0]->display_objects(
            objects         => \@moves,
            display_columns => [qw(id authuser efshost pid request)],
        );

        die qq{This command may not be used while active moves are in progress\n};

    }

    my $project_lock = EFS::DirLock->new({
        directory => q{/efs/dev/efs/config-namespace},
        retry     => 1,
        command   => $history->request,
    });

    my $fm_cells = EFS::ForkManager->new({

        list => \@cells,

        action => sub {

            my $cell = shift;

            return EFS::RPC->execute(
                cell   => $cell,
                class  => $class,
                method => q{rebuild_cell},
                args   => [
                    cell => $cell->name,
                    type => $args{type},
                ],
            );

        },

    });

    $fm_cells->run || die(
        qq{Unable to rebuild the $args{type} namespace in these cells:\n\t},
        join(qq{\n\t},map { $_->name } $fm_cells->get_failed), qq{\n},
    );

    return 1;

}

sub rebuild_cell {

    my $class = shift;

    my (%args) = validate( @_, {
        cell => { type => SCALAR },
        type => { type => SCALAR },
    });

    my $cell_rs = $class->get_resultset( q{Cell} );
    my $cell    = $cell_rs->retrieve_bynames_assert( name => $args{cell} );
    my $fstype  = $cell->fstype;
    my $type    = $args{type};

    my %rs_of = (
        dev  => $class->get_resultset( q{Project} ),
        home => $class->get_resultset( q{Home} ),
    );

    my $admin_path = $cell->admin_path( type => $type );

    if ( $fstype eq q{nfs} ) {
        if ( not $cell->supports_filetype( $type ) ) {
            STDOUT->print( qq{Cell $cell has no $type namespace. Skipping.\n} );
            return 1;
        }
        EFS::Util->MakeDir( path => qq{$admin_path/global} );
    }

    my @objects = $rs_of{ $type }->all;

    my %dirs  = ();
    my %links = ();

    foreach my $object ( @objects ) {

        my $object_id = $object->object_id;

        $dirs{ dirname($object_id) }++;

        if ( $type eq q{home} ) {
            $dirs{ dirname( dirname($object_id) ) }++;
        }

        $links{ $object_id } = $object->autopath;

    }

    my %changes = (
        dirs  => 0,
        links => 0,
    );

    given ( $fstype ) {

        when ( q{afs} ) {
            $changes{dirs} += EFS::AFS::Util->create_dirs(
                volume => qq{efs.global.$type},
                cell   => $cell->name,
                dirs   => [ sort keys %dirs ],
            );
            $changes{links} += EFS::AFS::Util->create_symlinks(
                volume => qq{efs.global.$type},
                cell   => $cell->name,
                links  => \%links,
            );
        }

        when ( q{nfs} ) {
            foreach my $dir ( sort keys %dirs ) {
                $changes{dirs} += EFS::Util->MakeDir(
                    path => qq{$admin_path/global/$dir}
                );
            }
            foreach my $from ( sort keys %links ) {
                $changes{links} += EFS::Util->MakeLink(
                    from => qq{$admin_path/global/$from},
                    to   => $links{$from},
                );
            }
        }

    }

    my %patterns = (
        home => qq{$admin_path/global/*/*/*},
        dev  => qq{$admin_path/global/*/*},
    );

    my @links = bsd_glob( $patterns{$type}, GLOB_ERR );
    my @dirs  = ();

    if ( GLOB_ERROR ) {
        die qq{Unable to glob list of $type symlinks from $admin_path: $ERRNO\n};
    }

    my %removes = (
        dirs  => 0,
        links => 0,
    );

    foreach my $path ( @links ) {

        ( my $link = $path ) =~ s{$admin_path/global/}{}msx;

        if ( not $links{$link} ) {
            EFS::Util->RemoveLink( from => $path );
            $removes{links}++;
        }

        my $dir = dirname $link;

        if ( not $dirs{$dir} ) {
            push @dirs, $dir;
        }

    }

    foreach my $dir ( @dirs ) {
        EFS::Util->RemoveDirHierarchy(
            empty => 1,
            top   => qq{$admin_path/global/$dir},
        );
        $removes{dirs}++;
    }

    if ( %changes || %removes ) {

        STDOUT->print(
            qq{Change summary for the $type namespace in $cell:\n},
        );

        if ( %changes ) {
            STDOUT->print(
                qq{\tlinks added or changed => $changes{links}\n},
                qq{\tdirs  added            => $changes{dirs}\n},
            );
        }

        if ( %removes ) {
            STDOUT->print(
                qq{\tlinks removed          => $removes{links}\n},
                qq{\tdirs  removed          => $removes{dirs}\n},
            );
        }

    } else {
        STDOUT->print( qq{No changes for $type namespace in $cell\n} );
    }

    return 1;

}

sub addentry {

    my $class = shift;

    my (%args) = validate( @_, {
        type     => { type => SCALAR },
        realpath => { type => SCALAR },
        autopath => { type => SCALAR },
        cells    => { type => ARRAYREF, optional => 1 },
    });

    my $type = $args{ type };

    if ( $type ne q{dev} && $type ne q{home} ) {
        die(
            q{Invalid argument.  } .
                qq{The only supported namespaces are 'dev' and 'home'\n},
        );
    }

    my $cell_rs = $class->get_resultset( q{Cell} );

    my @cells =
        $args{cells} ? @{ $args{cells} } :
            $cell_rs->search({ celltype => q{dev}, hidden => 0 });

    my $fm_cells = EFS::ForkManager->new({

        list => \@cells,

        action => sub {

            my $cell = shift;

            return EFS::RPC->execute(
                cell   => $cell,
                class  => $class,
                method => q{addentry_cell},
                args   => [
                    cell     => $cell->name,
                    type     => $args{type},
                    realpath => $args{realpath},
                    autopath => $args{autopath},
                ],
            );

        },

    });

    $fm_cells->run || die(
        qq{Unable to addentry to $args{type} namespace in these cells:\n\t},
        join( qq{\n\t}, $fm_cells->get_failed), qq{\n},
    );

    return 1;

}

sub addentry_cell {

    my $class = shift;
    my %args  = @_;

    my $cell_rs = $class->get_resultset( q{Cell} );
    my $cell = $cell_rs->retrieve_bynames_assert( name => $args{cell} );

    my $type     = $args{type};
    my $realpath = $args{realpath};
    my $autopath = $args{autopath};

    given ( $cell->fstype ) {

        when ( q{afs} ) {
            EFS::AFS::Util->create_symlink(
                volume => qq{efs.global.$type},
                cell   => $cell->name,
                path   => $realpath,
                target => $autopath,
            );
        }

        when ( q{nfs} ) {

            if ( not $cell->supports_filetype( $type ) ) {
                STDOUT->print( qq{Cell $cell has no $type namespace. Skipping.\n} );
                return 1;
            }

            my $admin_path = $cell->admin_path( type => $type );
            my $parentdir = dirname( $realpath );
            EFS::Util->MakeDirHierarchy(
                top    => qq{$admin_path/global},
                bottom => qq{$admin_path/global/$parentdir},
            );
            EFS::Util->MakeLink(
                from => qq{$admin_path/global/$realpath},
                to   => $autopath,
            );

        }

    }

    return 1;

}

sub delentry {

    my $class = shift;

    my (%args) = validate( @_, {
        type     => { type => SCALAR },
        realpath => { type => SCALAR },
        showpath => { type => SCALAR },
        cells    => { type => ARRAYREF, optional => 1 },
    });

    if ( $args{type} ne q{dev} && $args{type} ne q{home} ) {
        die(
            q{Invalid argument. } .
                qq{The only supported namespaces are 'dev' and 'home'\n},
        );
    }

    my $cell_rs = $class->get_resultset( q{Cell} );

    my @cells =
        $args{cells} ? @{ $args{cells} } :
            $cell_rs->search({ celltype => q{dev}, hidden => 0 });

    my $fm_cells = EFS::ForkManager->new({

        list => \@cells,

        action => sub {

            my $cell = shift;

            return EFS::RPC->execute(
                cell   => $cell,
                class  => $class,
                method => q{delentry_cell},
                args   => [
                    cell     => $cell->name,
                    type     => $args{type},
                    realpath => $args{realpath},
                    showpath => $args{showpath},
                ],
            );

        },

    });

    $fm_cells->run || die(
        qq{Unable to delentry to $args{type} namespace in these cells:\n\t},
        join(qq{\n\t},map { $_->name } $fm_cells->get_failed), qq{\n},
    );

    return 1;

}

sub delentry_cell {

    my $class = shift;
    my %args  = @_;

    my $cell_rs = $class->get_resultset( q{Cell} );
    my $cell = $cell_rs->retrieve_bynames_assert( name => $args{cell} );

    my $type     = $args{type};
    my $realpath = $args{realpath};

    given ( $cell->fstype ) {

        when ( q{afs} ) {
            EFS::AFS::Util->remove_symlink(
                volume => qq{efs.global.$type},
                cell   => $cell->name,
                path   => $realpath,
            );
        }

        when ( q{nfs} ) {
            if ( not $cell->supports_filetype( $type ) ) {
                STDOUT->print( qq{Cell $cell has no $type namespace. Skipping.\n} );
                return 1;
            }
            my $admin_path = $cell->admin_path( type => $type );
            EFS::Util->RemoveLink( from => qq{$admin_path/global/$realpath} );
        }

    }

    return 1;

}

sub assert_no_active_rebuilds {

    my $class = shift;

    my $history_rs = $class->get_resultset( q{History} );
    my @active = $history_rs->search({
        command => q{rebuild namespace},
        status  => undef,
    });

    if ( @active ) {

        STDOUT->print(
            qq{The following rebuild namespace commands are still active:\n},
        );

        $active[0]->display_objects(
            objects         => \@active,
            display_columns => [qw(id authuser efshost pid request)],
        );

        die(
            q{This command may not be used while } .
                qq{active rebuilds are in progress\n},
        );

    }

    return;

}

1;

# Local Variables:
#   mode: cperl
#   cperl-indent-level: 4
#   fill-column: 78
# End:
# vim: expandtab shiftwidth=4:
