package EFS::CLI::SSLServer;

use strict;
use warnings;
use English;

use base q{EFS::CLI};
use base q{EFS::DBI};

use EFS::Colors;

use Params::Validate qw( :all );
use Date::Format;
use File::Basename qw(basename);
use Date::Calc qw( Delta_Days Decode_Month );
use Readonly;
use Try::Tiny;

Readonly my $SCREEN_WIDTH => 1024;
Readonly my $DEFAULT_SERVER_CERTIFICATE_WARNING => 14;
Readonly my $DEFAULT_ROOT_CERTIFICATE_WARNING => 30;

my %templates = ();

$templates{root} = <<'END_OF_ROOT_TEMPLATE';
[ ca ]
default_ca              = efs_ca

[ efs_ca ]
dir                     = {$base_dir}
certificate             = $dir/cacert.pem
database                = $dir/index.txt
new_certs_dir           = $dir/certs
private_key             = $dir/cakey.pem
serial                  = $dir/serial

default_crl_days        = 7
default_days            = {$default_days}
default_md              = sha1
unique_subject          = no

policy                  = efs_ca_policy
x509_extensions         = certificate_extensions

[ efs_ca_policy ]
commonName              = supplied
stateOrProvinceName     = supplied
countryName             = supplied
emailAddress            = supplied
organizationName        = supplied
organizationalUnitName  = optional

[ certificate_extensions ]
basicConstraints        = CA:false

[ req ]
default_bits            = 2048
default_keyfile         = {$base_dir}/cakey.pem
default_md              = sha1

prompt                  = no
distinguished_name      = root_ca_distinguished_name

x509_extensions         = root_ca_extensions

[ root_ca_distinguished_name ]
commonName              = {$servername}
stateOrProvinceName     = AnyState
countryName             = US
emailAddress            = efsops@{$root_domain}
organizationName        = EFS Root Certificate Authority

[ root_ca_extensions ]
basicConstraints        = CA:true
END_OF_ROOT_TEMPLATE

$templates{server} = <<'END_OF_SERVER_TEMPLATE';
[ req ]
default_bits            = 2048
default_keyfile         = {$base_dir}/cakey.pem
default_md              = sha1

prompt                  = no
distinguished_name      = server_ca_distinguished_name

x509_extensions         = server_ca_extensions

[ server_ca_distinguished_name ]
commonName              = {$servername}
stateOrProvinceName     = AnyState
countryName             = US
emailAddress            = efsops@{$root_domain}
organizationName        = EFS Server

[ server_ca_extensions ]
basicConstraints        = CA:true

END_OF_SERVER_TEMPLATE

sub define {

    my $class = shift;

    my (%args) = validate( @_, {
        %EFS::CLI::validate_common,
        type      => { type => SCALAR },
        efsserver => { type => SCALAR },
    });

    my $efsserver_rs = $class->get_resultset(q{EFSServer});
    my $efsserver = $efsserver_rs->retrieve_bynames_assert( name => $args{efsserver} );

    if ( $args{type} ne q{primary} and $args{type} ne q{secondary} ) {
        die qq{Invalid argument: -type must be either primary or secondary\n};
    }

    my $resultset = $class->get_resultset;

    my %bytype = map { $_->type            => $_ } $resultset->all;
    my %byname = map { $_->efsserver->name => $_ } $resultset->all;

    if ( $bytype{ $args{type} } ) {

        if ( $bytype{ $args{type} }->efsserver->name eq $args{efsserver} ) {
            STDOUT->print(
                qq{EFSServer $args{efsserver} is already defined as },
                qq{the $args{type} SSLServer\n},
            );
            return 1;
        }

        if ( $byname{ $args{efsserver} } ) {
            my $type = $byname{ $args{efsserver} }->type;
            die(
                qq{EFSServer $args{efsserver} is already defined as },
                qq{the $type SSLServer\n},
            );
        }

        my $sslserver = $bytype{ $args{type} };

        $sslserver->set_column( efsserver => $efsserver->id );
        $sslserver->update;

        STDOUT->print( qq{Defined $args{efsserver} as the $args{type} SSLServer\n} );

    }
    elsif ( $byname{ $args{efsserver} } ) {

        my $type = $byname{ $args{efsserver} }->type;
        die(
            qq{EFSServer $args{efsserver} is already defined as },
            qq{the $type SSLServer\n},
        );

    }
    else {

        my $sslserver = $resultset->create({
            type      => $args{type},
            efsserver => $efsserver->id,
        });

        STDOUT->print( qq{Defined $args{efsserver} as the $args{type} SSLServer\n} );

    }

    return 1;

}

sub undef {
    return die(
        qq{SSLServer may not be undefined\n},
        qq{There should always be two: a primary and secondary\n},
    );
}

sub swaptypes {

    my $class = shift;

    my $resultset = $class->get_resultset;

    my ($primary)   = $resultset->search({ type => q{primary} });
    my ($secondary) = $resultset->search({ type => q{secondary} });

    if ( not $primary or not $secondary ) {
        die(
            qq{Both the primary and secondary SSLServers must be defined\n},
            qq{in order to swap their types\n},
        );
    }

    $primary->set_column( type => q{temporary} );
    $primary->update;

    $secondary->set_column( type => q{primary} );
    $secondary->update;

    $primary->set_column( type => q{secondary} );
    $primary->update;

    STDOUT->print(
        qq{The primary and secondary SSLServers have swapped types\n},
    );

    return 1;

}

sub history {

    my $class = shift;

    my (%args) = validate( @_, {
        %EFS::CLI::validate_common,
        %EFS::CLI::History::validation_specs,
    });

    return EFS::CLI::History->history(%args);

}

sub initca {

    my $class = shift;

    my (%args) = validate( @_, {
        %EFS::CLI::validate_common,
        type     => { type => SCALAR },
        force    => { type => SCALAR, optional => 1 },
        name     => { type => SCALAR, optional => 1 },
    });

    my $resultset = $class->get_resultset;
    my ($primary) = $resultset->search({ type => q{primary} });

    if ( not $primary ) {
        die qq{No primary SSLServer has been defined yet.\n};
    }

    if ( not $primary->efsserver->is_localhost ) {
        die( q{_EFS_REDIRECT_SERVER: } . $primary->efsserver->name . qq{\n} );
    }

    my $parameter_rs = $class->get_resultset(q{Parameter});

    my $root_domain = $parameter_rs->configured(q{root_domain}) ||
        die qq{No root_domain parameter has been defined.\n};

    my $root_lifetime =
        $parameter_rs->configured(q{ssl_root_certificate_lifetime}) ||
            die qq{No root certificate lifetime has been defined.\n};

    my $server_lifetime =
        $parameter_rs->configured(q{ssl_server_certificate_lifetime}) ||
            die qq{No server certificate lifetime has been defined.\n};

    if ( $args{type} eq q{primary} ) {

        $class->initca_primary(
            %args,
            primary         => $primary,
            root_domain     => $root_domain,
            root_lifetime   => $root_lifetime,
            server_lifetime => $server_lifetime,
        );

    }
    elsif ( $args{type} eq q{secondary} ) {

        my ($secondary) = $resultset->search({ type => q{secondary} });

        if ( not $secondary ) {
            die qq{No secondary SSLServer has been defined yet.\n};
        }

        $class->initca_secondary(
            %args,
            secondary       => $secondary,
            root_domain     => $root_domain,
            root_lifetime   => $root_lifetime,
            server_lifetime => $server_lifetime,
        );

    }
    else {
        die qq{Invalid argument: -type must be either primary or secondary\n};
    }

    return 1;

}

sub initca_primary {

    my $class = shift;
    my %args  = @_;

    my $primary     = $args{primary};
    my $root_domain = $args{root_domain};

    my $root_lifetime   = $args{root_lifetime};
    my $server_lifetime = $args{server_lifetime};

    my $caroot = q{/var/efs/ca};

    if ( -d $caroot ) {
        if ( $args{force} ) {
            my $backup = time2str( qq{$caroot-%Y%m%d-%H%M%S}, time, q{GMT} );
            rename ( $caroot, $backup ) ||
                die qq{Unable to backup existing $caroot: $ERRNO\n};
            STDOUT->print( qq{Existing $caroot renamed to $backup\n} );
        }
        else {
            die(
                qq{The primary SSLServer has already been initialized.\n},
                qq{To force reinitialization, use -force, but...\n},
                qq{This will require creating new SSL server certificates\n},
                qq{for ALL existing EFSServers.\n},
                qq{You have been warned.  Proceed with caution.\n},
            );
        }
    }

    STDOUT->print( qq{Initializing primary SSLServer\n} );

    EFS::Util->MakeDir( path => $caroot, mode => 0700 );

    my $dirlock = EFS::DirLock->new({
        directory => $caroot,
        retry     => 1,
        command   => qq{$class->initssl_primary},
    });

    EFS::Util->MakeDir( path => qq{$caroot/certs}, mode => 0700 );
    EFS::Util->MakeDir( path => qq{$caroot/reqs},  mode => 0700 );

    system (qq{echo 01 > $caroot/serial});
    $CHILD_ERROR and die qq{Unable to initialize $caroot/serial\n};

    system (qq{touch $caroot/index.txt});
    $CHILD_ERROR and die qq{Unable to initialize $caroot/index.txt\n};

    $class->create_openssl_config(
        type         => q{root},
        file         => qq{$caroot/openssl.conf},
        base_dir     => $caroot,
        servername   => $primary->efsserver->name,
        root_domain  => $root_domain,
        default_days => $server_lifetime,
    );

    $class->create_passphrase( dir => $caroot );

    $class->create_root_certificate(
        caroot        => $caroot,
        root_lifetime => $root_lifetime,
    );

    $class->create_crl( caroot => $caroot );

    return 1;

}

sub initca_secondary {
    return die qq{Not sure how to implement the secondary SSLServer yet\n};
}

sub initssl_efsserver {

    my $class     = shift;
    my $efsserver = shift;

    my $servername = $efsserver->name;

    my $parameter_rs = $class->get_resultset(q{Parameter});

    my $root_domain = $parameter_rs->configured(q{root_domain}) ||
        die qq{No root_domain parameter has been defined.\n};

    my $caroot  = q{/var/efs/ca};
    my $sslroot = q{/var/efs/ssl};

    my $dirlock = EFS::DirLock->new({
        directory => $caroot,
        retry     => 1,
        command   => qq{$class->initssl_server},
    });

    my $date           = time2str( q{%Y%m%d-%H%M%S}, time, q{GMT} );
    my $sslroot_tmp    = qq{$sslroot-$servername-$date};
    my $sslroot_remote = qq{$sslroot-$date};

    STDOUT->print( qq{Initializing SSL on EFSServer $servername\n} );

    EFS::Util->MakeDir( path => $sslroot_tmp, mode => 0700 );

    $class->create_openssl_config(
        type        => q{server},
        file        => qq{$sslroot_tmp/openssl.conf},
        base_dir    => $sslroot,
        servername  => $servername,
        root_domain => $root_domain,
    );

    $class->create_passphrase( dir => $sslroot_tmp );

    EFS::CLI::EFSServer->create_certificate_request(
        config  => qq{$sslroot_tmp/openssl.conf},
        keyout  => qq{$sslroot_tmp/$servername-key.pem},
        out     => qq{$sslroot_tmp/$servername-req.pem},
        passout => qq{file:$sslroot_tmp/passphrase},
    );

    $class->create_server_certificate(
        config => qq{$caroot/openssl.conf},
        in     => qq{$sslroot_tmp/$servername-req.pem},
        out    => qq{$caroot/certs/$servername-cert.pem},
    );

    system (qq{cp $caroot/certs/$servername-cert.pem $sslroot_tmp});
    $CHILD_ERROR and die qq{Unable to copy server certificate to $sslroot_tmp\n};

    system (qq{cp $caroot/cacert.pem $sslroot_tmp});
    $CHILD_ERROR and die qq{Unable to copy root certificate to $sslroot_tmp\n};

    system (qq{cp $caroot/crl.pem $sslroot_tmp});
    $CHILD_ERROR and die qq{Unable to copy CRL list to $sslroot_tmp\n};

    EFS::Rsync->new({
        dest_efsserver => $efsserver,
        src_path       => $sslroot_tmp,
        dest_path      => $sslroot_remote,
    });

    my $ssh = q{ssh -o StrictHostKeyChecking=no -o LogLevel=ERROR};

    my $identity = EFS::CLI::EFSServer->sshid;

    if ( -f $identity ) {
        $ssh .= qq{ -i $identity};
    }
    elsif ( not $ENV{EFS_TEST_SUITE} ) {
        warn( qq{No ssh identify file $identity\n}, "Continuing...\n" );
    }

    system qq{$ssh $servername mv $sslroot_remote/cacert.pem /var/efs/etc/cacert.pem};
    $CHILD_ERROR and die qq{Unable to install cacert.pem on $servername\n};

    system qq{$ssh $servername mv $sslroot_remote/openssl.conf /var/efs/etc/openssl.conf};
    $CHILD_ERROR and die qq{Unable to install openssl.conf on $servername\n};

    system qq{$ssh $servername mv $sslroot_remote/crl.pem /var/efs/etc/crl.pem};
    $CHILD_ERROR and die qq{Unable to install crl.pem on $servername\n};

    system qq{$ssh $servername rm -f $sslroot};
    $CHILD_ERROR and die qq{Unable to remove $sslroot symlink on $servername\n};

    my $sslroot_name = basename($sslroot_remote);

    system qq{$ssh $servername ln -s $sslroot_name $sslroot};
    $CHILD_ERROR and die qq{Unable to create $sslroot symlink on $servername\n};

    EFS::Util->RemoveDirHierarchy( top => $sslroot_tmp );

    return 1;

}

sub create_passphrase {

    my $class = shift;
    my %args  = @_;

    my $dir = $args{dir};

    my $passphrase = qq{$dir/passphrase};

    my $io = IO::File->new(qq{>$passphrase}) ||
        die qq{Unable to open $passphrase: $ERRNO\n};

    $io->print(
        Text::Password::Pronounceable->generate( 20, 30 ), qq{\n},
        Text::Password::Pronounceable->generate( 20, 30 ), qq{\n},
    ) || die qq{Unable to write to $passphrase: $ERRNO\n};

    $io->close || die qq{Unable to close $passphrase: $ERRNO\n};

    return 1;

}

sub create_root_certificate {

    my $class = shift;
    my %args  = @_;

    my $caroot = $args{caroot};

    my $openssl_cmd = join(
        q{},
        qq{openssl req -config $caroot/openssl.conf },
        q{-x509 -newkey rsa:2048 },
        qq{-days $args{root_lifetime} },
        qq{-out $caroot/cacert.pem -outform PEM },
        qq{-keyout $caroot/cakey-tmp.pem },
        qq{-passout file:$caroot/passphrase },
        q{2>&1}
    );

    my $openssl_out = qx{ $openssl_cmd };
    $CHILD_ERROR and die qq{Unable to create self-signed root certificate:\n$openssl_out};

    $openssl_cmd = join(
        q{},
        q{openssl rsa },
        qq{-in $caroot/cakey-tmp.pem },
        qq{-out $caroot/cakey.pem },
        qq{-passin file:$caroot/passphrase },
        q{2>&1}
    );

    $openssl_out = qx{ $openssl_cmd };
    $CHILD_ERROR and die qq{Unable to remove passphrase from root certificate:\n$openssl_out};

    unlink( qq{$caroot/cakey-tmp.pem} ) ||
        die qq{Unable to unlink $caroot/cakey-tmp.pem: $ERRNO\n};

    STDOUT->print( qq{Created self-signed root certificate\n} );

    return 1;

}

sub create_server_certificate {

    my $class = shift;
    my %args  = @_;

    my $openssl_cmd = join(
        q{},
        qq{openssl ca -config $args{config} },
        qq{-in $args{in} -out $args{out} },
        q{-batch 2>&1 },
    );
    my $openssl_out = qx{ $openssl_cmd };
    $CHILD_ERROR and die qq{Unable to issue new server certificate:\n $openssl_out};

    STDOUT->print( qq{Issued new server certificate: $args{out}\n} );

    return 1;

}

sub create_openssl_config {

    my $class = shift;
    my %args  = @_;

    my $template = Text::Template->new(
        TYPE   => q{STRING},
        SOURCE => $templates{ $args{type} },
    );

    if ( not $template ) {
        die qq{Unable to create $args{type} template: $Text::Template::ERROR\n};
    }

    my $contents = $template->fill_in(
        HASH => {
            base_dir     => $args{base_dir},
            root_domain  => $args{root_domain},
            servername   => $args{servername},
            default_days => $args{default_days},
        },
    );

    my $io = IO::File->new(qq{>$args{file}}) ||
        die qq{Unable to open $args{file}: $ERRNO\n};

    $io->print($contents) ||
        die qq{Unable to write to $args{file}: $ERRNO\n};

    $io->close ||
        die qq{Unable to close $args{file}: $ERRNO\n};

    return 1;

}

sub showcerts {

    my $class = shift;

    my (%args) = validate( @_, {
        %EFS::CLI::validate_common,
    });

    my $resultset = $class->get_resultset;
    my ($primary) = $resultset->search( { type => q{primary} } );

    if ( not $primary ) {
        die qq{No primary SSLServer has been defined yet.\n};
    }

    if ( not $primary->efsserver->is_localhost ) {
        die( q{_EFS_REDIRECT_SERVER: } . $primary->efsserver->name . qq{\n});
    }

    my $parameter_rs = $class->get_resultset( q{Parameter} );

    my $server_certificate_warning =
        $parameter_rs->configured( q{ssl_server_certificate_warning} ) ||
            $DEFAULT_SERVER_CERTIFICATE_WARNING;

    my $root_certificate_warning =
        $parameter_rs->configured( q{ssl_root_certificate_warning} ) ||
            $DEFAULT_ROOT_CERTIFICATE_WARNING;

    my $caroot = q{/var/efs/ca};

    my $openssl_out = qx{ openssl x509 -in $caroot/cacert.pem -text -noout 2>&1 };
    $CHILD_ERROR and die qq{Unable to read root certificate:\b$openssl_out};

    my %root = ();

    foreach my $line ( split m{\n+}msx, $openssl_out ) {

        if ( $line =~ m{Not\s+Before:\s+(.*$)}msx ) {
            $root{issued} = $1;
        }

        if ( $line =~ m{Not\s+After\s+:\s+(.*$)}msx ) {
            $root{expires} = $1;
        }

    }

    if ( not $root{expires} ) {
        die qq{Unable to determine expiration date of root certificate\n};
    }

    $root{days_left} = $class->days_left( $root{expires} );

    my $in = IO::File->new(qq{$caroot/index.txt}) ||
        die qq{Unable to open $caroot/index.txt: $ERRNO\n};

    my %active  = ();
    my %revoked = ();

    while ( defined ( my $line = $in->getline ) ) {

        my ( $expiration, $revocation, $serial, $dn, $name );

        if ( $line =~ m/^V\s+/msx ) {
            ( $expiration, $serial, $dn )
                = ( split ( m/\s+/msx, $line, 5 ) )[ 1, 2, 4 ];
        }
        elsif ( $line =~ m/^R\s+/msx ) {
            ( $expiration, $revocation, $serial, $dn ) =
                ( split ( m/\s+/msx, $line, 6 ) )[ 1, 2, 3, 5 ];
        }

        my $days_left = $class->days_left($expiration);

        if ( $dn =~ m{/CN=(.*?)/}msx ) {
            $name = $1;
        }
        else {
            die qq{Unable to parse $caroot/index.txt line:\n$line};
        }

        if ( $revocation ) {
            $revoked{$serial} = {
                expiration => $expiration,
                recovation => $revocation,
                name       => $name,
                days_left  => $days_left,
            };
        }
        else {
            $active{$serial} = {
                expiration => $expiration,
                name       => $name,
                days_left  => $days_left,
            };
        }

    }

    my $table_root = Text::FormatTable->new(q{l  l  r});
    $table_root->head(qw(Issued Expires Days_Left));
    $table_root->rule(q{=});

    my @root = (
        $class->normalize_date( $root{issued} ),
        $class->normalize_date( $root{expires} ),
        $root{days_left},
    );

    if ( $root{days_left} <= 0 ) {
        $table_root->row( map { color_bad($_) } @root );
    }
    elsif ( $root{days_left} <= $root_certificate_warning ) {
        $table_root->row( map { color_warn($_) } @root );
    }
    else {
        $table_root->row( @root );
    }

    my $table_active = Text::FormatTable->new(q{r  l  l  r});
    $table_active->head(qw(Serial EFSServer Expires Days_Left));
    $table_active->rule(q{=});

    foreach my $serial ( sort { $a <=> $b } keys %active ) {

        my @row = (
            $serial,
            $active{$serial}->{name},
            $class->normalize_date($active{$serial}->{expiration}),
            $active{$serial}->{days_left},
        );

        if ( $active{$serial}->{days_left} <= 0 ) {
            $table_active->row( map { color_bad($_) } @row );
        }
        elsif ( $active{$serial}->{days_left} <= $server_certificate_warning ) {
            $table_active->row( map { color_warn($_) } @row );
        }
        else {
            $table_active->row( @row );
        }

    }

    my $table_revoked = Text::FormatTable->new(q{r  l  l  l  r});
    $table_revoked->head(qw(Serial EFSServer Expires Revoked Days_Left));
    $table_revoked->rule(q{=});

    foreach my $serial ( sort { $a <=> $b } keys %revoked ) {

        my @row = (
            $serial,
            $revoked{$serial}->{name},
            $class->normalize_date($revoked{$serial}->{expiration}),
            $class->normalize_date($revoked{$serial}->{recovation}),
            $revoked{$serial}->{days_left},
        );

        if ( $revoked{$serial}->{days_left} <= 0 ) {
            $table_revoked->row( map { color_warn($_) } @row );
        }
        else {
            $table_revoked->row( @row );
        }

    }

    STDOUT->print(
        qq{Root Certificate\n},
        $table_root->render($SCREEN_WIDTH)
    );

    if ( %active ) {
        STDOUT->print(
            qq{Active Server Certificates\n},
            $table_active->render($SCREEN_WIDTH)
        );
    }
    else {
        STDOUT->print(qq{\nNo active server certificates\n\n});
    }

    if ( %revoked ) {
        STDOUT->print(
            qq{Revoked Server Certificates\n},
            $table_revoked->render($SCREEN_WIDTH)
        );
    }
    else {
        STDOUT->print(qq{\nNo revoked server certificates\n\n});
    }

    return 1;

}

sub purgecerts_sslserver {

    my $class = shift;
    my %names = @_;

    my $caroot = q{/var/efs/ca};

    my $dirlock = EFS::DirLock->new({
        directory => $caroot,
        retry     => 1,
        command   => qq{$class->purgecerts_sslserver},
    });

    my $in = IO::File->new(qq{$caroot/index.txt}) ||
        die(qq{Unable to open $caroot/index.txt: $ERRNO\n});

    my $out = IO::File->new(qq{>$caroot/index.txt.new}) ||
        die(qq{Unable to open $caroot/index.txt.new: $ERRNO\n});

    my %serial_of = ();

    while ( defined ( my $line = $in->getline ) ) {

        my ( $expiration, $serial, $dn );

        if ( $line =~ m/^V\s+/msx ) {
            ( $expiration, $serial, $dn )
                = ( split ( m/\s+/msx, $line, 5 ) )[ 1, 2, 4 ];
        }
        elsif ( $line =~ m/^R\s+/msx ) {
            ( $expiration, $serial ) = ( split ( m/\s+/msx, $line ) )[ 1, 3 ];
        }

        if ( $class->days_left($expiration) <= 0 ) {

            unlink (qq{$caroot/certs/$serial.pem}) ||
                die(qq{Unable to unlink $caroot/certs/$serial.pem: $ERRNO\n});

            STDOUT->print(
                qq{Purged expired cert $caroot/certs/$serial.pem\n}
            );

            next;

        }

        if ( $line =~ m/^V\s+/msx ) {

            if ( $dn =~ m{/CN=(.*?)/}msx ) {
                push @{ $serial_of{$1} }, $serial;
            }
            else {
                die qq{Unable to parse $caroot/index.txt line:\n$line};
            }

        }

        $out->print($line) ||
            die qq{Unable to write to $caroot/index.txt.new: $ERRNO\n};

    }

    $in->close ||
        die qq{Unable to close $caroot/index.txt: $ERRNO\n};

    $out->close ||
        die qq{Unable to close $caroot/index.txt.new: $ERRNO\n};

    rename( qq{$caroot/index.txt.new}, "$caroot/index.txt" ) ||
        die qq{Unable to rename $caroot/index.txt.new -> index.txt: $ERRNO\n};

    foreach my $name ( sort keys %names ) {

        if ( not $serial_of{$name} ) {
            warn qq{No certificates found for $name\n};
            next;
        }

        my @serial = @{ $serial_of{$name} };
        pop @serial;    # Most recent is the valid one

        foreach my $serial (@serial) {
            $class->revoke_certificate(
                caroot => $caroot,
                cert   => qq{$caroot/certs/$serial.pem},
            );
            STDOUT->print( qq{Revoked old certificate for $name ($serial.pem)\n} );
        }

    }

    $class->create_crl( caroot => $caroot );

    $class->publish_crl( sort keys %names );

    STDOUT->print( qq{Expired and/or purged old certificates\n} );

    return 1;

}

sub publish_crl {

    my $class = shift;
    my @names = @_;

    my $fm_efsservers = EFS::ForkManager->new({

        list => \@names,

        action => sub {

            my $name = shift;

            my $efsserver_rs = $class->get_resultset(q{EFSServer});
            my ($efsserver) = $efsserver_rs->search({ name => $name });

            EFS::Rsync->new({
                dest_efsserver => $efsserver,
                src_path       => q{/var/efs/ca/crl.pem},
                dest_path      => q{/var/efs/etc/crl.pem},
            });

        },

    });

    $fm_efsservers->run || die(
        qq{Unable to publish CRL list on:\n\t},
        join ( qq{\n\t}, $fm_efsservers->get_failed ), qq{\n}
    );

    STDOUT->print( qq{Published new CRL list\n} );

    return 1;

}

sub create_crl {

    my $class = shift;
    my %args  = @_;

    my $caroot = $args{caroot};

    my $openssl_cmd = join(
        q{},
        qq{openssl ca -config $caroot/openssl.conf },
        qq{-gencrl -out $caroot/crl.pem 2>&1},
    );

    my $openssl_out = qx{ $openssl_cmd };
    $CHILD_ERROR and die qq{Unable to create $caroot/crl.pem:\n$openssl_out};

    STDOUT->print( qq{Created new certificate revocation list\n} );

    return 1;

}

sub revoke_certificate {

    my $class = shift;
    my (%args) = @_;

    my $caroot = $args{caroot};

    my $openssl_cmd = join(
        q{},
        qq{openssl ca -config $caroot/openssl.conf },
        qq{-revoke $args{cert} 2>&1},
    );

    my $openssl_out = qx{ $openssl_cmd };
    $CHILD_ERROR and die qq{Unable to revoke certificate:\n$openssl_out};

    return 1;

}

sub normalize_date {

    my $class = shift;
    my $date = shift;

    if ( my ($year,$month,$day,$hour,$min,$sec)  =
             ( $date =~
                   m{^(\d{2})(\d{2})(\d{2})(\d{2})(\d{2})(\d{2})Z$}msx ) ) {
        $year = qq{20$year};
        return qq{$year-$month-$day $hour:$min:$sec};
    }

    if ( my ($month,$day,$hour,$min,$sec,$year)  =
             ( $date =~
                   m{^(\S+)\s+(\S+)\s+(\d{2}):(\d{2}):(\d{2})\s+(\d{4})}msx ) ) {
        $month = Decode_Month($month);
        return qq{$year-$month-$day $hour:$min:$sec};
    }

    die qq{Unable to parse date '$date'\n};

    return;

}

sub days_left {

    my $class      = shift;
    my $expiration = shift;

    my ( $eyear, $emonth, $eday );

    if ( $expiration =~ m{^(\d{2})(\d{2})(\d{2})\d{6}Z$}msx ) {
        ( $eyear, $emonth, $eday ) = ( $1, $2, $3 );
        $eyear = qq{20$eyear};
    }
    elsif ( $expiration =~
                m{^(\S+)\s+(\S+)\s+\d{2}:\d{2}:\d{2}\s+(\d{4})}msx ) {
        ( $eyear, $emonth, $eday ) = ( $3, Decode_Month($1), $2 );
    }
    else {
        die qq{Unable to parse expiration $expiration\n};
    }

    my ( $nyear, $nmonth, $nday ) = ( localtime (time) )[ 5, 4, 3 ];
    $nyear += 1900;
    $nmonth += 1;

    my $delta;

    try {
        $delta = Delta_Days( $nyear, $nmonth, $nday, $eyear, $emonth, $eday );
    }
    catch {
        die qq{Unable to calculate expiration delta:\n$_};
    };

    return $delta;

}

1;

# Local Variables:
#   mode: cperl
#   cperl-indent-level: 4
#   fill-column: 78
# End:
# vim: expandtab shiftwidth=4:
