# Abstract class for items that are distable--Any, Metaproj, Project,
# Release, ReleaseAlias

package EFS::DistableObject;

use 5.010;
use strict;
use warnings;
use English qw(-no_match_vars);

use Params::Validate qw( :all );
use List::MoreUtils qw( any all );
use Readonly;

### CLI class methods

sub diststate {

    my $class = shift;

    my ($self,$type) = $class->get_distable_object( \@_ );

    my (%args) = validate( @_, {
        verbose => { type => BOOLEAN | UNDEF, optional => 1 },
        %EFS::CLI::validate_diststate_args,
    });

    if ( $type ne q{EFS} ) {

        # For efficiency's sake, we'll do sanity checks here, since
        # Any doesn't need them

        if ( $self->can(q{installs}) and not $self->installs->count ) {
            die qq{This $type has no installs, and thus has no diststate\n};
        }

        # XXX: need to change regex if celltype can be a list
        if ( not $self->has_prod_stages
                 and ( $args{celltype} eq q{prod}
                     or $args{celltype} eq q{all} ) ) {

            # We display a different message depending on whether
            # the item has releases below it (i.e., is a Metaproj or Project)
            # or not
            if ( $self->can(q{releases}) ) {
                STDOUT->print(
                    qq{NOTE: \u$type $self has no prod or deprecated\n},
                    qq{      releases, and therefore has no 'prod' diststate.\n},
                );
            }
            else {
                STDOUT->print(
                    qq{NOTE: \u$type $self has stage },
                    $self->stage, qq{\n},
                    qq{      and therefore has no 'prod' diststate.\n},
                );
            }

            # XXX: need to change if celltype can be a list
            $args{celltype} = q{dev};
        }
    }

    my @cells = EFS::DBI->handle_cell_arguments(
        %args,
        default_scope => q{global},
    );

    my $diststate = $self->get_diststate;

    # If CLI provided -cells, we don't re-sort the cells for the table
    $diststate->display_text_table(
        cells                 => \@cells,
        verbose               => $args{verbose},
        respect_cell_ordering => exists $args{cells},
    );

    return 1;
}

sub list {

    my $class = shift;

    my ($self,$type) = $class->get_distable_object( \@_ );

    my (%args) = validate( @_, {
        all       => { type => BOOLEAN, optional => 1 },
        devpaths  => { type => BOOLEAN, optional => 1 },
        showlinks => { type => BOOLEAN, optional => 1 },
    });

    if ( $args{all} and $args{devpaths} ) {
        die qq{Options -all and -devpaths cannot be used together.\n};
    }

SUBITEM:
    foreach my $subitem ( sort $self->subitems ) {

        my $destroyed = $subitem->is_destroyed;

        if ( not $args{all} and $destroyed ) {
            next SUBITEM;
        }

        my $line
            = $args{devpaths} ? $subitem->get_path(q{dev}) : $subitem->name;

        if ( $args{showlinks} ) {
            if ( $subitem->isa(q{EFS::DBI::Result::ReleaseAlias}) ) {
                $line .= q{@ -> } . $subitem->target->name;
            }

            if ($destroyed) {
                $line .= q{ (destroyed)};
            }
        }

        STDOUT->print( qq{$line\n} );

    }

    return;

}

sub set_owner_list {
    my ($class,@args) = @_;
    my ($self,$type) = $class->get_distable_object(\@args);
    return $self->_set_notification_list( list => q{owner_list}, @args);
}

sub set_notify_list {
    my ($class,@args) = @_;
    my ($self,$type) = $class->get_distable_object(\@args);
    return $self->_set_notification_list( list => q{notify_list}, @args);
}

sub clear_owner_list {
    my ($class,@args ) = @_;
    my ($self,$type) = $class->get_distable_object(\@args);
    return $self->_set_notification_list( list => q{owner_list}, clear => 1, @args);
}

sub clear_notify_list {
    my ($class,@args) = @_;
    my ($self,$type) = $class->get_distable_object(\@args);
    return $self->_set_notification_list( list => q{notify_list}, clear => 1, @args);
}

sub get_owner_list {
    my ($class,@args) = @_;
    my ($self,$type) = $class->get_distable_object(\@args);
    return $self->_get_notification_list( list => q{owner_list} );
}

sub get_notify_list {
    my ($class,@args) = @_;
    my ($self,$type) = $class->get_distable_object(\@args);
    return $self->_get_notification_list( list => q{notify_list} );
}

sub test_owner_list {
    my ($class,@args) = @_;
    my ($self,$type) = $class->get_distable_object(\@args);
    return $self->_test_notification_list( list => q{owner_list}, @args );
}

sub test_notify_list {
    my ($class,@args) = @_;
    my ($self,$type) = $class->get_distable_object(\@args);
    return $self->_test_notification_list( list => q{notify_list}, @args );
}

### Object methods

Readonly our @PROD_STAGES     => qw( prod deprecated );
Readonly our @REJECTED_STAGES => qw( rejected deprecated );

sub has_prod_stages {
    return shift->_has_stages_matching( sub { $_ ~~ @PROD_STAGES } );
}

sub has_only_prod_stages {
    return shift->_has_only_stages_matching( sub { $_ ~~ @PROD_STAGES } );
}

sub has_only_destroyable_stages {
    return shift->_has_only_stages_matching( sub { $_ ~~ @REJECTED_STAGES } );
}

sub _has_stages_matching {
    my ($self,$condition) = @_;
    my @releases = $self->stage_relevant_releases;
    return any { $condition->() } map { $_->stage } @releases;
}

sub _has_only_stages_matching {
    my ($self,$condition) = @_;
    my @releases = $self->stage_relevant_releases;
    return all { $condition->() } map { $_->stage } @releases;
}

# For a releasealias, the stage comes from the target release; for
# a install or release, the release; for anything else, it'll be
# an amalgam of the releases inside the any/metaproj/project
sub stage_relevant_releases {
    my $self = shift;
    return $self->first_matching_method(
        qw(target_release release releases)
    );
}

sub first_matching_method {

    my ($self,@meths) = @_;

    my @meths_found;

    foreach my $meth (@meths) {

        if ( $self->can($meth) ) {
            push @meths_found, $meth;
            my @results = $self->$meth;
            if ( not @results ) {
                warn(
                    sprintf(q{INTERNAL ERROR: Object '%s' of class %s }
                                . qq{supports\n  the method %s, but returned no}
                                    . qq{ results.\n},
                        $self, ref ($self), $meth )

                );
            }
            elsif ( any { not defined } @results ) {
                my @defined = grep {defined} @results;
                my @undefed = grep { not defined } @results;
                warn(
                    sprintf(q{INTERNAL ERROR: Object '%s' of class %s }
                                . q{supports the method %s, but returned}
                                    . q{ %d undef results (+%d defined ones). },
                        $self,
                        ref ($self),
                        $meth,
                        scalar @undefed,
                        scalar @defined
                    )
                );
            }
            else {
                return @results;
            }
        }

    }

    if (@meths_found) {
        die(
            sprintf (
                qq{INTERNAL ERROR: Object '%s' of class %s supports\n}
                        . q{the methods %s, but none of them returned validly.},
                $self,
                ref ($self),
                join ( q{, }, @meths_found )
            )
        );
    }
    else {
        die(
            sprintf(q{INTERNAL ERROR: Object '%s' of class %s does not }
                        . qq{support\n  any of the methods %s.},
                $self,
                ref ($self),
                    join ( q{, }, @meths )
            )
        );
    }
    return;
}

# Return the project or metaproject's owner list
sub mp_owner_list {

    my $self = shift;

    if ( $self->can(q{owner_list}) ) {
        return $self->owner_list;
    }
    elsif ( $self->can(q{project}) ) {
        return $self->project->owner_list || $self->metaproj->owner_list;
    }
    elsif ( $self->can(q{metaproj}) ) {
        return $self->metaproj->owner_list;
    }
    else {
        return;
    }

}

# CLI methods here can use this to get the object, regardless of whether
# it's any, metaproj, project, release or releasealias
# Call it like this:
#     my ($self, $type) = $class->get_distable_object(\@_);
# before you validate the args, and it will remove the object args.
#
# For any distable object, $type is the result of class2type.  For
# Any, that's 'EFS'.
sub get_distable_object {

    my $class   = shift;
    my $arg_ref = shift;

    my (%args) = validate_with(
        params => $arg_ref,
        spec   => {
            metaproj => { type => SCALAR | OBJECT, optional => 1 },
            project  => { type => SCALAR | OBJECT, optional => 1 },
            name     => { type => SCALAR | OBJECT, optional => 1 },
        },
        allow_extra => 1,
    );

    my ($self,$type);

    if ( exists $args{name} ) {

        # We're working with something other than Any

        my %retrieve_args;
        for my $arg (qw(metaproj project name)) {
            if ( exists $args{$arg} ) {
                $retrieve_args{$arg} = $args{$arg};
                delete $args{$arg};
            }
        }

        $self = $class->retrieve_bynames_assert(
            %retrieve_args,
            disabled => 1,
        );

        $type = $self->class2type;
    }
    else {
        # We're working with any, so "object" is actually "class"
        $self = $class;
        $type = q{EFS};
    }

    # Remove object spec args from @_
    @{$arg_ref} = (%args);

    return ($self,$type);

}

# For things with the concept of "destroyed", return it.
# For anything else, return false.
sub is_destroyed {
    my $self = shift;
    if ( $self->can(q{destroyed}) ) {
        return $self->destroyed;
    }
    return;
}

# Concatenate attributes in this object and all its superobjects
# These are NOT uniquified, because it may be useful to have duplicates
# in some future context--user of function should uniq it if necessary
sub all_mpr_attrs {

    my ($self,@attrs) = @_;

    my @results;
    if ( $self->container ) {
        push @results, $self->container->all_mpr_attrs(@attrs);
    }

    for my $attr (@attrs) {
        if ( defined ( my $value = $self->get_attribute( name => $attr ) ) ) {
            push @results, $value;
        }
    }

    return @results;
}

# Same thing for a table column
sub all_mpr_fields {

    my ($self,@fields) = @_;

    my @results;
    if ( $self->container ) {
        push @results, $self->container->all_mpr_fields(@fields);
    }

    for my $field (@fields) {
        if ( $self->can($field) ) {
            if ( defined ( my $value = $self->$field ) ) {
                push @results, $value;
            }
        }
    }

    return @results;

}

sub _set_notification_list {

    my ($self,@args) = @_;

    my (%args) = validate_with(
        params => \@args,
        spec   => {
            address => {
                regex    => qr/\A (?: \S+ \@ \S+ [.] \S+ | ) \z/xms,
                optional => 1,
            },
            list  => { regex => qr/\A owner_list | notify_list \z/xms, },
            clear => { type => BOOLEAN, default => 0, },
        },
        allow_extra => 1
    );

    if ( not $args{clear} and not $args{address} ) {
        X::AddressEmpty->throw;
    }

    $self->set_column(
        $args{list} => ( $args{address} ? $args{address} : undef )
    );

    return $self->update;

}

sub _get_notification_list {

    my ($self,@args) = @_;

    my (%args) = validate( @args, {
        list => { regex => qr/\A owner_list | notify_list \z/xms, },
    });

    if ( my $list = $self->get_column( $args{list} ) ) {
        STDOUT->print(qq{$list\n});
        return 1;
    }
    else {
        my $container = $self;
        while ( $container = $container->container ) {
            if ( my $list = $container->get_column( $args{list} ) ) {
                STDOUT->print(qq{$list (Inherited from $container)\n});
                return 1;
            }
        }
    }

    warn qq{$self has no $args{list} set.\n};

    return;

}

sub _test_notification_list {

    my ($self,@args) = @_;

    my (%args) = validate_with(
        params => \@args,
        spec => {
            list     => { regex => qr/\A owner_list | notify_list \z/xms, },
            username => 1,
        },
        allow_extra => 1,
    );

    my %notify_args;
    if ( $args{list} eq q{notify_list} ) {
        $notify_args{send_to_owner_list} = 0;
    }
    elsif ( $args{list} eq q{owner_list} ) {
        $notify_args{send_to_notify_list} = 0;
    }

    EFS::Notification->notify_mprs_template(
        %args,
        %notify_args,
        template      => q{notify_test},
        template_args => {
            list        => $args{list},
            object      => $self->stringify_self,
            object_type => $self->class2type,
            user        => $args{authuser},
        },
        reply_to   => $self,
        recipients => [$self],
    );
    STDOUT->print(qq{Mail sent.\n});
    return;
}

sub assert_uploadable {

    my $self = shift;

    my $type = ucfirst $self->class2type;

    my $object = $self;

    if ( $self->isa( q{EFS::DBI::Result::Install} ) ||
             $self->isa( q{EFS::DBI::Result::Release} ) ||
                 $self->isa( q{EFS::DBI::Result::ReleaseAlias} ) ) {
        $object = $self->project;
    }

    my $upload_server   = $object->find_mpr_attr( q{upload_server} );
    my $upload_username = $object->find_mpr_attr( q{upload_username} );
    my $upload_sshkey   = $object->find_mpr_attr( q{upload_sshkey} );
    my $upload_root     = $object->find_mpr_attr( q{upload_root} );

    if ( not $upload_server or not $upload_username or
             not $upload_sshkey or not $upload_root ) {
        die(
            qq{$type $self is not configured for upload\n},
            qq{All 4 of the following attributes must be defined:\n},
            qq{\tupload_server\n},
            qq{\tupload_username\n},
            qq{\tupload_sshkey\n},
            qq{\tupload_root\n},
            qq{for either the metaproj, or the project\n},
        );
    }

    return 1;

}

sub assert_downloadable {

    my $self = shift;

    my $type = ucfirst $self->class2type;

    my $object = $self;

    if ( $self->isa( q{EFS::DBI::Result::Install} ) ||
             $self->isa( q{EFS::DBI::Result::Release} ) ||
                 $self->isa( q{EFS::DBI::Result::ReleaseAlias} ) ) {
        $object = $self->project;
    }

    my $download_url = $object->find_mpr_attr( q{download_url} );

    if ( not $download_url ) {
        die(
            qq{$type $self is not configured for download\n},
            qq{The following attribute must be defined:\n},
            qq{\tdownload_url\n},
            qq{for either the metaproj, or the project\n},
        );
    }

    return 1;

}

sub get_afsvolume {

    my $self = shift;
    my $type = shift;

    my $method = qq{${type}volume};

    my $afsvolume = $self->$method;

    if ( not $afsvolume ) {

        my $afsvolume_rs = $self->get_resultset( q{AFSVolume} );

        $afsvolume = $afsvolume_rs->register(
            type   => $type,
            parent => $self,
        );

        $self->set_column( $method => $afsvolume->id );
        $self->update;

    }

    return $afsvolume;
    
}

1;

# Local Variables:
#   mode: cperl
#   cperl-indent-level: 4
#   fill-column: 78
# End:
# vim: expandtab shiftwidth=4:
