package EFS::Diststate::Install;

use strict;
use warnings;
use English;
use Carp;

use Moose;
use Params::Validate qw( :all );
extends qw(EFS::Diststate);
use Set::Scalar;

has install => (
    is => 'rw',
);

has updated => (
    is => 'rw',
    default => sub { {} },
);



sub BUILD {
    my ($self, @args) = @_;

    my (%args) = validate( @args, {
        install => {
            type => OBJECT,
            isa  => q{EFS::DBI::Result::Install},
        },
    });

    my $cell_rs = EFS::DBI->get_resultset( q{Cell} );

    my $install = $self->install( $args{install} );

    my @cells = ();

    if ( $install->has_prod_stages() ) {
        @cells      = $cell_rs->all;
    } else {
        @cells      = (
            $cell_rs->search({ celltype => q{dev} }),
            $cell_rs->search({ celltype => q{test} }),
        );
    }

    for my $cell ( @cells ) {

        my $updated = $install->updatetime_for_cell($cell);

        if ( $updated ) {
            $self->updated->{$cell} = $updated;
        } else {
            $self->updated->{$cell} = 0;
        }

    }

    return 1;

}

sub get_checktime {
    return shift->install->checktime;
}

sub related_item { return shift->install }

sub needs_update {

    my $self = shift;

    my (%args) = validate( @_, {
        cells => { type => ARRAYREF, optional => 1 },
    });

    if ( $self->cells_needing_update(%args) ) {
        return 1;
    }
    else {
        return;
    }
}

sub installs_needing_update {

    my $self = shift;

    my (%args) = validate( @_, {
        cells => { type => ARRAYREF, optional => 1 },
    });

    if ( $self->cells_needing_update(%args) ) {
        return $self->install;
    }
    else {
        return;
    }

}

sub cells_needing_update {

    my $self  = shift;

    my (%args) = validate( @_, {
        cells => { type => ARRAYREF, optional => 1 },
    });

    return $self->_gather_cells_matching(
        filter => sub {
            # we want the cells where the checktime is newer than the
            # update time
            $self->get_checktime > $self->updated->{ $_[0] };
        },
        %args
    );
}

sub has_been_updated {

    my $self  = shift;

    my (%args) = validate( @_, {
        cells => { type => ARRAYREF, optional => 1 },
    });

    # If cells argument not given, we use all known cells
    my @wanted_cells = ref($args{cells}) ? @{ $args{cells} } : ();
    if ( not @wanted_cells ) {
        @wanted_cells = keys %{ $self->updated };
    }

    # We need to be sure *all* cells in %args have been updated
    my $wanted_cellset =
        Set::Scalar->new( map { ref $_ ? $_->name : $_ } @wanted_cells );
    my $updated_cellset =
        Set::Scalar->new( map { $_->name } $self->cells_updated(%args) );

    if ( $wanted_cellset == $updated_cellset ) {
        return 1;
    } else {
        return;
    }

}

sub installs_updated {

    my $self  = shift;

    my (%args) = validate( @_, {
        cells => { type => ARRAYREF, optional => 1 },
    });

    if ( $self->has_been_updated(%args) ) {
        return $self;
    }
    else {
        return;
    }

}

sub cells_updated {

    my $self  = shift;

    my (%args) = validate( @_, {
        cells => { type => ARRAYREF, optional => 1 },
    });

    return $self->_gather_cells_matching(
        filter => sub {
            # we want the cells where there exists a (non-zero) update
            # time
            $self->updated->{ $_[0] };
        },
        %args
    );

}

# For an install, the concept of partial update degenerates to
# just the list of cells updated, so we make
# cells_partially_updated an alias for cells_updated
{
    no strict q{refs}; ## no critic (ProhibitNoStrict)
    *{cells_partially_updated} = \&cells_updated;
}


# Factors out identical code from needs_update and has_been_updated
sub _gather_cells_matching {

    my $self  = shift;

    my (%args) = validate( @_, {
        cells  => { type => ARRAYREF, optional => 1 },
        filter => { type => CODEREF, },
    });

    my $cell_rs = EFS::DBI->get_schema->resultset( q{Cell} );

    my %interested_in = map { $_ => $_ } @{ $args{cells} };

    my @results = ();

    for my $cell_name ( keys %{ $self->updated } ) {
        if ( @{ $args{cells} } ) {  # if option not given, check all cells
            next unless $interested_in{$cell_name};
        }
        if ( $args{filter}->($cell_name) ) {
            push @results,
                $cell_rs->retrieve_bynames_assert( name => $cell_name );
        }
    }

    return @results;
}


1;

# Local Variables:
#   mode: cperl
#   cperl-indent-level: 4
#   fill-column: 78
# End:
# vim: expandtab shiftwidth=4:
