package EFS::ExecLinks;

use strict;
use warnings;
use English;

use File::Basename qw( dirname basename );
use Params::Validate qw( :all );

sub get_all_release_data {

    my $class = shift;

    my $metaproj_attrs_rs = EFS::DBI->get_resultset( q{MetaprojAttrs} );
    my $project_attrs_rs  = EFS::DBI->get_resultset( q{ProjectAttrs} );
    my $release_attrs_rs  = EFS::DBI->get_resultset( q{ReleaseAttrs} );
    my $release_rs        = EFS::DBI->get_resultset( q{Release} );

    my $data = {
        metaproj => $metaproj_attrs_rs->get_flattened_data,
        project  => $project_attrs_rs->get_flattened_data,
        release  => $release_attrs_rs->get_flattened_data,
    };

    my @releases = $release_rs->search({ locked => 1 });

    foreach my $release ( @releases ) {
        $data->{release}->{$release->object_id}->{locked} = 1;
    }

    return $data;

}

sub get_release_attr {

    my $class = shift;
    my %args = @_;

    my $data      = $args{data};
    my $attribute = $args{attribute};
    my $release   = $args{release};
    my $project   = dirname $release;
    my $metaproj  = dirname $project;

    my $value =
        $data->{metaproj}->{$metaproj}->{$attribute} ||
            $data->{project}->{$project}->{$attribute} ||
                $data->{release}->{$release}->{$attribute};

    return $value;

}

sub get_all_platform_data {

    my $class = shift;

    my $platform_rs       = EFS::DBI->get_resultset( q{Platform} );
    my $instance_rs       = EFS::DBI->get_resultset( q{PlatformInstance} );
    my $platform_alias_rs = EFS::DBI->get_resultset( q{PlatformAlias} );
    my $instance_alias_rs = EFS::DBI->get_resultset( q{PlatformInstanceAlias} );

    my $data = {
        platform              => $platform_rs->get_flattened_data,
        platforminstance      => $instance_rs->get_flattened_data,
        platformalias         => $platform_alias_rs->get_flattened_data,
        platforminstancealias => $instance_alias_rs->get_flattened_data,
    };

    foreach my $type ( keys %{ $data } ) {
        foreach my $name ( keys %{ $data->{$type} } ) {
            $data->{names}->{$name} = $type;
        }
    }

    $data->{hierarchy} = $platform_rs->get_hierarchy_order;

    return $data;

}

sub get_aliases_of {

    my $class = shift;
    my %args = @_;

    my $type = $args{type};
    my $data = $args{data};
    my $name = $args{name};

    my @aliases = ();

    foreach my $alias ( keys %{ $data->{ $type } } ) {
        if ( $data->{ $type }->{ $alias } eq $name ) {
            push @aliases, $alias;
        }
    }

    return @aliases;

}

sub get_instances_of {

    my $class = shift;
    my %args = @_;

    my $type = $args{type};
    my $data = $args{data};
    my $name = $args{name};

    my @instances = ();

    foreach my $instance ( keys %{ $data->{platforminstance} } ) {
        if ( $data->{platforminstance}->{$instance}->{platform} eq $name ) {
            push @instances, $instance;
        }
    }

    return @instances;

}

sub get_forwards_of {

    my $class = shift;
    my %args = @_;

    my $type = $args{type};
    my $data = $args{data};
    my $name = $args{name};

    my @forwards = ();

    foreach my $forwards ( keys %{ $data->{$type} } ) {
        if ( $data->{$type}->{$forwards}->{backwards} eq $name ) {
            push @forwards, $forwards;
        }
    }

    return @forwards;

}

sub get_emulators_of {

    my $class = shift;
    my %args = @_;

    my $type = $args{type};
    my $data = $args{data};
    my $name = $args{name};

    my @emulators = ();

    foreach my $emulator ( keys %{ $data->{$type} } ) {
        if ( $data->{$type}->{$emulator}->{emulates} eq $name ) {
            push @emulators, $emulator;
        }
    }

    return @emulators;

}

sub update_links {

    my $class = shift;

    my (%args) = validate( @_, {
        directory              => { type     => SCALAR },
        disable_exec_backwards => { type     => SCALAR | UNDEF,
                                    optional => 1 },
        disable_exec_emulates  => { type     => SCALAR | UNDEF,
                                    optional => 1 },
        platform_data          => { type     => HASHREF,
                                    optional => 1 },
        command                => { type     => SCALAR | UNDEF,
                                    optional => 1 },
        locked                 => { type     => SCALAR | UNDEF,
                                    optional => 1 },
        makedir                => { type     => HASHREF,
                                    optional => 1 },
    });

    my $directory = $args{directory};

    my ($type,$admin_path,$cell,$release) =
        $class->parse_exec_dir( $directory );

    my $lock;

    if ( $args{command} ) {
        $lock = EFS::DirLock->new({
            directory => $args{directory},
            retry     => 1,
            command   => $args{command},
        });
    }

    my $io = IO::Dir->new($directory) ||
        die qq{Unable to opendir $directory: $ERRNO\n};

    my $platform_data = $args{platform_data} || $class->get_all_platform_data;

    my %dirs = ();
    my %links = ();

    foreach my $name ( $io->read ) {

        next if $name ~~ [qw(. ..)];

        if ( -l qq{$directory/$name} ) {
            $links{$name}++;
            next;
        }

        if ( -d qq{$directory/$name} ) {
            $dirs{$name}++;
            next;
        }

    }

    if ( $args{makedir} ) {
        my $name = delete $args{makedir}->{name};
        $args{makedir}->{path} = qq{$directory/$name};
        EFS::Util->MakeDir( %{ $args{makedir} } );
        $dirs{$name}++;
        delete $links{$name};
    }

    my $exec_links = $class->exec_links(
        dirs                   => \%dirs,
        platform_data          => $platform_data,
        disable_exec_backwards => $args{disable_exec_backwards},
        disable_exec_emulates  => $args{disable_exec_emulates},
    );

    foreach my $link ( sort keys %{ $exec_links->{links} } ) {
        # If locked, adds only, no changes
        if ( not $args{locked} or not -l qq{$directory/$link}) {
            EFS::Util->MakeLink(
                from => qq{$directory/$link},
                to   => $exec_links->{links}->{$link},
            );
        }
    }

    foreach my $link ( sort keys %links ) {
        if ( not $args{locked} and not $exec_links->{links}->{$link} ) {
            EFS::Util->RemoveLink( from => qq{$directory/$link} );
        }
    }

    # create_instance_exec_links

    my $dotexecpath = $type eq q{dev} ? q{/install/.exec} : q{/.exec};
    my $execpath    = $type eq q{dev} ? q{/install/exec}  : q{/exec};

    my @names = ( @{ $exec_links->{instances} },
                  @{ $exec_links->{platforms} } );

    foreach my $name ( sort @names ) {

        my $from = qq{$admin_path/$name/$type/$release/$execpath};
        my $to   = qq{/efs/$type/$release/$dotexecpath/$name};

        my $from_base = dirname $from;

        if ( not -d $from_base ) {
            EFS::Util->MakeDirHierarchy(
                top    => qq{$admin_path/$name},
                bottom => $from_base,
            );
        }

        if ( not $args{locked} or not -l $from ) {
            EFS::Util->MakeLink(
                from => $from,
                to   => $to,
            );
        }

    }

    # create exec symlink

    EFS::Util->MakeLink(
        from => qq{$admin_path/$type/$release/$execpath},
        to   => qq{/.efs/$type/$release/$execpath},
    );

    return 1;

}

sub exec_links {

    my $class = shift;

    my (%args) = validate( @_, {
        dirs                   => { type     => HASHREF },
        platform_data          => { type     => HASHREF },
        disable_exec_backwards => { type     => SCALAR | UNDEF,
                                    optional => 1 },
        disable_exec_emulates  => { type     => SCALAR | UNDEF,
                                    optional => 1 },
    });

    my $exec_links = {
        dirs      => $args{dirs},
        links     => {},
        seen      => {},
        instances => [],
        platforms => [],
    };

    my $platform_data = $args{platform_data} || $class->get_all_platform_data;

    foreach my $name ( @{ $platform_data->{hierarchy} } ) {

        next unless exists $exec_links->{dirs}->{$name};

        $exec_links = $class->expand_links(
            name                   => $name,
            exec_links             => $exec_links,
            disable_exec_backwards => $args{disable_exec_backwards},
            disable_exec_emulates  => $args{disable_exec_emulates},
            platform_data          => $platform_data,
        );

    }

    delete $exec_links->{seen};

    foreach my $from ( keys %{ $exec_links->{dirs} } ) {
        if ( $platform_data->{ platforminstance }->{ $from } ) {
            push @{ $exec_links->{instances} }, $from;
        }
        if ( $platform_data->{ platform }->{ $from } ) {
            push @{ $exec_links->{platforms} }, $from;
        }
    }

    foreach my $from ( keys %{ $exec_links->{links} } ) {
        if ( $platform_data->{ platforminstance }->{ $from } ) {
            push @{ $exec_links->{instances} }, $from;
        }
        if ( $platform_data->{ platform }->{ $from } ) {
            push @{ $exec_links->{platforms} }, $from;
        }
        my $to = $exec_links->{links}->{$from};
        while ( $exec_links->{links}->{$to} ) {
            $to = $exec_links->{links}->{$to};
        }
        $exec_links->{flattened}->{$from} = $to;
    }

    $exec_links->{links} = delete $exec_links->{flattened};

    return $exec_links;

}

sub expand_links {

    my $class = shift;

    my (%args) = validate( @_, {
        name                   => { type     => SCALAR },
        exec_links             => { type     => HASHREF },
        disable_exec_backwards => { type     => SCALAR | UNDEF,
                                    optional => 1 },
        disable_exec_emulates  => { type     => SCALAR | UNDEF,
                                    optional => 1 },
        platform_data          => { type     => HASHREF },
    });

    my $exec_links = $args{exec_links};
    my $name       = $args{name};

    if ( $exec_links->{seen}->{$name} ) {
        return $exec_links;
    }

    $exec_links->{seen}->{$name}++;

    my $platform_data = $args{platform_data};

    my $type = $platform_data->{ names }->{ $name } ||
        die qq{No such platform(instance) or alias '$name'\n};

    if ( $type =~ s{alias$}{}msx ) {
        my $newname =
            $platform_data->{platformalias}->{$name} ||
                $platform_data->{platforminstancealias}->{$name};
        if ( $exec_links->{dirs}->{$newname} ) {
            return $exec_links;
        } else {
            $exec_links->{links}->{$newname} = $name;
            $name = $newname;
        }
    } else {
        my @aliases = $class->get_aliases_of(
            data => $platform_data,
            type => qq{${type}alias},
            name => $name,
        );
        foreach my $alias ( @aliases ) {
            if ( not exists $exec_links->{dirs}->{$alias} ) {
                $exec_links->{links}->{$alias} = $name;
            }
        }
    }

    my @children = ();

    if ( $type eq q{platform} ) {
        push @children, $class->get_instances_of(
            data => $platform_data,
            name => $name,
        );
    }

    if ( not $args{disable_exec_emulates} ) {
        push @children, $class->get_emulators_of(
            data => $platform_data,
            type => $type,
            name => $name,
        );
    }

    if ( not $args{disable_exec_backwards} ) {
        push @children, $class->get_forwards_of(
            data => $platform_data,
            type => $type,
            name => $name,
        );
    }

    @children = sort {
        exists $exec_links->{dirs}->{$b} <=> exists $exec_links->{dirs}->{$a}
    } @children;

    foreach my $child ( @children ) {

        if ( not exists $exec_links->{dirs}->{$child} ) {
            $exec_links->{links}->{$child} ||= $name;
        }

        $exec_links = $class->expand_links(
            name                   => $child,
            exec_links             => $exec_links,
            disable_exec_backwards => $args{disable_exec_backwards},
            disable_exec_emulates  => $args{disable_exec_emulates},
            platform_data          => $platform_data,
        );

    }

    return $exec_links;

}

sub parse_exec_dir {

    my $class = shift;
    my $dir = shift;

    my $type = $dir =~ m{ /install/.exec $ }msx ? q{dev} : q{dist};

    ## no critic ProhibitComplexRegexes
    my ($admin_path,$cell,$release) =
        ( $dir =~ m{
                       (
                           /.efsadmin/
                           ( [^/]+ ) # cell
                           .*
                       ) # admin_path
                       /$type/
                       (
                           [^/]+ / [^/]+ / [^/]+
                       ) # m/p/r
                       ( /install )?
                       /.exec
                       $
               }msx );

    if ( not $admin_path or not $cell or not $release ) {
        die qq{Unable to parse exec dir: $dir\n};
    }

    return ($type,$admin_path,$cell,$release);

}

1;

# Local Variables:
#   mode: cperl
#   cperl-indent-level: 4
#   fill-column: 78
# End:
# vim: expandtab shiftwidth=4:
