#!/usr/bin/env perl

use strict;
use warnings;
use English;

use Test::More tests => 101;

use lib q{test/lib};
use EFS::Test;

my $test = EFS::Test->new;
my $config = $test->get_config;

my $ldc_name      = $config->get_value( q{LocalDevCell}, q{name} );
my $ldc_shortname = $config->get_value( q{LocalDevCell}, q{shortname} );

my $rdc_name      = $config->get_value( q{RemoteDevCell}, q{name} );
my $rdc_shortname = $config->get_value( q{RemoteDevCell}, q{shortname} );

my $mp_name = q{testmeta testdisttree};
my $mp_path = q{testmeta/testdisttree};

my @releases = qw( a b c d e );

my @releasealiases = ( [qw(bl b)], );

# 1 2  3 4
# a bl b c
#   d  e
#   e

# 4 => c
# 3 => b e
# 2 => bl d
# 1 => a

my @depends = ( [qw(a bl)], [qw(b c)], [qw(d e)], [qw(a d)], [qw(a e)], );

my @setup = (
    q{efs display metaproj testmeta},
    qq{efs create project $mp_name $ldc_name},
);

foreach my $setup (@setup) {
    $test->efs_test_command(
        run_as  => q{efsops},
        name    => qq{(setup) $setup},
        command => $setup,
        bail    => 1,
    );
}

# Setup everything...
create_releases(@releases);
create_releasealiases(@releasealiases);
setup_dependencies( \@depends, q{define} );

# Now we actually start testing disttree

# These two commands should produce the SAME output, since they will
# be disting the SAME tree object.

my @commands = (
    qq{efs disttree release $mp_name disttree-a --global},
    qq{efs disttree project $mp_name --global},
);

foreach my $command (@commands) {

    update_releases(@releases);
    checknolock_releases(@releases);

    $test->efs_test_command(
        run_as  => q{efsops},
        command => qq{$command -dryrun},
        expect  => [
            qr{Stage \s+ Depth \s+ Diststate \s+ Runtime \s+ Dependencies}xms,
            qr{^=*}msx,
            qr{dev \s+ 1 \s+ 2d \s+ $mp_path/disttree-a}xms,
            qr{dev \s+ 2 \s+ 2d \s+ $mp_path/disttree-bl}xms,
            qr{dev \s+ 3 \s+ 2d \s+ $mp_path/disttree-b}xms,
            qr{dev \s+ 4 \s+ 2d \s+ $mp_path/disttree-c}xms,
            qr{dev \s+ 2 \s+ 2d \s+ $mp_path/disttree-d}xms,
            qr{dev \s+ 3 \s+ 2d \s+ $mp_path/disttree-e}xms,
            qr{dev \s+ 2 \s+ 2d \s+ $mp_path/disttree-e \s+ \(repeat\)}xms,
        ],
    );

    $test->efs_test_command(
        run_as    => q{efsops},
        command   => $command,
        filternot => qr{ ^ (Created|Changed) \s+ (directory|ownership) }msx,
        expect    => [
            q{Distributing 1 MPR at depth 4:},
            qq{$mp_path/disttree-c},
            dist_release_output(q{c}),
            q{Distributing 2 MPRs at depth 3:},
            qq{$mp_path/disttree-b},
            qq{$mp_path/disttree-e},
            dist_release_output(q{b}),
            dist_release_output(q{e}),
            q{Distributing 2 MPRs at depth 2:},
            qq{$mp_path/disttree-bl},
            qq{$mp_path/disttree-d},
            dist_releasealias_output( q{bl}, q{b} ),
            dist_release_output(q{d}),
            q{Distributing 1 MPR at depth 1:},
            qq{$mp_path/disttree-a},
            dist_release_output(q{a}),
            q{Successfully disted the following MPRs:},
            qq{$mp_path/disttree-a},
            qq{$mp_path/disttree-b},
            qq{$mp_path/disttree-bl},
            qq{$mp_path/disttree-c},
            qq{$mp_path/disttree-d},
            qq{$mp_path/disttree-e},
        ],
    );

    if ( $command =~ /disttree \s+ release/msx ) {
        setup_dependencies( [ [qw(a bl)], ], q{undef} );
        destroy_releasealiases(@releasealiases);
        create_releasealiases(@releasealiases);
        setup_dependencies( [ [qw(a bl)], ], q{define} );
    }

}

# Now test the failure scenario, but just for one of them
# Also, test the diststate check by disting one of the releases

update_releases(@releases);
checknolock_releases(@releases);

dist_releases(q{c});
update_releases(q{e});

$test->efs_test_command(
    run_as    => q{efsops},
    command   => qq{efs disttree release $mp_name disttree-a --global},
    not       => 1,
    filternot => qr{ ^ (Created|Changed) \s+ (directory|ownership) }msx,
    expect    => [
        q{Skipping 1 MPR at depth 4, which is up to date},
        q{Nothing left to distribute at depth 4},
        q{Distributing 2 MPRs at depth 3:},
        qq{$mp_path/disttree-b},
        qq{$mp_path/disttree-e},
        dist_release_output(q{b}),
        dist_release_failure(q{e}),
        q{Skipping 1 MPR at depth 2, due to dependent failures},
        q{Skipping 1 MPR at depth 2, which is up to date},
        q{Nothing left to distribute at depth 2},
        q{Skipping 1 MPR at depth 1, due to dependent failures},
        q{Nothing left to distribute at depth 1},
        q{Skipped the following up to date MPRs:},
        qq{$mp_path/disttree-bl},
        qq{$mp_path/disttree-c},
        q{Successfully disted the following MPRs:},
        qq{$mp_path/disttree-b},
        q{Failed to dist the following MPRs or their dependencies:},
        qq{$mp_path/disttree-a},
        qq{$mp_path/disttree-d},
        qq{$mp_path/disttree-e},
    ],
);

checknolock_releases(qw( e ));

$test->efs_test_command(
    run_as    => q{efsops},
    filternot => qr{ ^ (Created|Changed) \s+ (directory|ownership) }msx,
    command   => qq{efs disttree release $mp_name disttree-e --global},
    expect    => [
        q{Distributing 1 MPR at depth 1:},
        qq{$mp_path/disttree-e},
        dist_release_output(q{e}),
        q{Successfully disted the following MPRs:},
        qq{$mp_path/disttree-e},
    ],
);

update_releases(qw( d e ));
checkpoint_releases(qw( d e ));

$test->efs_test_command(
    run_as  => q{efsops},
    command => qq{efs disttree release $mp_name disttree-d --celltype prod --global},
    not     => 1,
    expect  => [
        q{Unable to distribute this dependency tree to the specified prod cells.},
        q{The following release(link)s have invalid stages:},
        qq{$mp_path/disttree-d => dev},
    ],
);

setstage_prod(qw( e d ));

$test->efs_test_command(
    run_as  => q{efsops},
    command => qq{efs disttree release $mp_name disttree-d --celltype prod --global},
);

# NOTE: This is a little wierd...
# Update c, d and e, and then disttree project.
# If the code mistakenly tries to dist c, it will fail.
# It should ONLY try to dist the prod releases

unlock_releases(qw( d e ));
update_releases(qw( c d e ));
checkpoint_releases(qw( d e ));

$test->efs_test_command(
    run_as  => q{efsops},
    command => qq{efs disttree project $mp_name --celltype prod --global},
);

# Make sure that we got child history records for the stuff we didn't
# dist explicitly (disttree-b and disttree-bl had NO explicit commands
# run above).
$test->efs_test_command(
    command => qq{efs history release $mp_name disttree-b -count 5},
    filter  => qr{^ \s+ \d+}msx,
    expect  => [
        qr{dist \s+ release \s+ testmeta \s+ testdisttree \s+ disttree-b}msx,
        qr{checkpoint \s+ release \s+ testmeta \s+ testdisttree \s+ disttree-b}msx,
        qr{dist \s+ release \s+ testmeta \s+ testdisttree \s+ disttree-b}msx,
        qr{checkpoint \s+ release \s+ testmeta \s+ testdisttree \s+ disttree-b}msx,
        qr{dist \s+ release \s+ testmeta \s+ testdisttree \s+ disttree-b}msx,
    ],
);

$test->efs_test_command(
    command => qq{efs history releasealias $mp_name disttree-bl -count 5},
    filter  => qr{^ \s+ \d+}msx,
    expect  => [
        qr{create \s+ releasealias \s+ testmeta \s+ testdisttree \s+ disttree-bl \s+ disttree-b}msx,
        qr{dist \s+ releasealias \s+ testmeta \s+ testdisttree \s+ disttree-bl}msx,
        qr{destroy \s+ releasealias \s+ testmeta \s+ testdisttree \s+ disttree-bl}msx,
        qr{create \s+ releasealias \s+ testmeta \s+ testdisttree \s+ disttree-bl \s+ disttree-b}msx,
        qr{dist \s+ releasealias \s+ testmeta \s+ testdisttree \s+ disttree-bl}msx,
    ],
);

# Teardown everything...
setup_dependencies( \@depends, q{undef} );
destroy_releasealiases(@releasealiases);
destroy_releases(@releases);

my @teardown = ( qq{efs destroy project $mp_name} );

foreach my $teardown (@teardown) {
    $test->efs_test_command(
        run_as  => q{efsops},
        name    => qq{(teardown) $teardown},
        command => $teardown,
    );
}

exit 0;

sub setup_dependencies {

    my ( $depends, $type ) = @_;

    foreach my $fromto (@{$depends}) {
        my ( $from, $to ) = @{$fromto};
        $test->efs_test_command(
            run_as  => q{efsops},
            command =>
                qq{efs $type dependency $mp_name disttree-$from } .
                    qq{$mp_name disttree-$to runtime},
        );
    }

    return;

}

sub create_releases {
    foreach my $release (@_) {
        $test->efs_test_command(
            run_as  => q{efsops},
            command => qq{efs create release $mp_name disttree-$release},
        );
        $test->efs_test_command(
            run_as  => q{efsops},
            command => qq{efs create install $mp_name disttree-$release common},
        );
    }
    return;
}

sub create_releasealiases {
    foreach my $fromto (@_) {
        my ( $from, $to ) = @{$fromto};
        $test->efs_test_command(
            run_as  => q{efsops},
            command => qq{efs create releasealias $mp_name disttree-$from disttree-$to},
        );
    }
    return;
}

sub destroy_releasealiases {
    foreach my $fromto (@_) {
        my ( $from, $to ) = @{$fromto};
        $test->efs_test_command(
            run_as  => q{efsops},
            command => qq{efs destroy releasealias $mp_name disttree-$from},
        );
    }
    return;
}

sub unlock_releases {
    foreach my $release (@_) {
        $test->efs_test_command(
            run_as  => q{efsops},
            command => qq{efs unlock release $mp_name disttree-$release},
        );
    }
    return;
}

sub update_releases {
    foreach my $release (@_) {
        $test->efs_test_command(
            command => qq{touch /efs/dev/$mp_path/disttree-$release/install/common/foo},
        );
    }
    return;
}

sub checknolock_releases {
    foreach my $release (@_) {
        $test->efs_test_command(
            run_as  => q{efsops},
            command => qq{efs checkpoint release $mp_name disttree-$release -nolock},
        );
    }
    return;
}

sub checkpoint_releases {
    foreach my $release (@_) {
        $test->efs_test_command(
            run_as  => q{efsops},
            command => qq{efs checkpoint release $mp_name disttree-$release},
        );
    }
    return;
}

sub dist_releases {
    foreach my $release (@_) {
        $test->efs_test_command(
            run_as  => q{efsops},
            command => qq{efs dist release $mp_name disttree-$release --global},
        );
    }
    return;
}

sub setstage_prod {
    foreach my $release (@_) {
        $test->efs_test_command(
            run_as  => q{efsops},
            command => qq{efs setstage release $mp_name disttree-$release prod},
        );
    }
    return;
}

sub destroy_releases {
    foreach my $release (@_) {

        my $stage = q{rejected};

        if ( grep { $release eq $_ } qw( d e ) ) {
            $stage = q{deprecated};
        }

        $test->efs_test_command(
            run_as  => q{efsops},
            command => qq{efs setstage release $mp_name disttree-$release $stage},
        );
        $test->efs_test_command(
            run_as  => q{efsops},
            command => qq{efs destroy release $mp_name disttree-$release},
        );
        $test->efs_test_command(
            run_as  => q{efsops},
            command => qq{efs purge release $mp_name disttree-$release -now},
        );

    }
    return;
}

sub dist_release_output {

    my $release = shift;

    my @output = (
        q{Dev environment diststate:},
        qr{Install \s+ $ldc_shortname \s+ $rdc_shortname}xms,
        qr{^=+}msx,
        qr{common \s+ \w \s+ \w}xms,
        qr{^-+}msx,
        q{Legend:},
        q{D: Item is disted and up to date},
        q{P: Item is disted but needs update},
        q{X: Item has not been disted},
        qq{Validating the checkpoint state for release '$mp_path/disttree-$release'.},
    );

    if ( -l qq{/efs/dist/$mp_path/disttree-$release} ) {
        push @output,
            qq{Removed symlink $ldc_name:/efs/dist/$mp_path/disttree-$release};
    }

    push @output, (
        qq{Distributing $mp_path/disttree-$release from $ldc_name to $ldc_name},
        q{Distribution completed:},
        qq{Target Path = $ldc_name:/efs/dist/$mp_path/disttree-$release/common},
        q{Transfered 1 of 2 total files, 0 of 0 total bytes},
        q{Paths updated:},
        q{foo},
        qr{Elapsed \s+ time \s+ = \s+ \d+ \s+ secs}msx,
        qq{Successfully distributed $mp_path/disttree-$release/common to $ldc_name},
        qq{Successfully distributed the following installs to cell $ldc_name:},
        qq{$mp_path/disttree-$release/common},
        qq{Successfully distributed all installs to $ldc_name},
        qq{Distributing $mp_path/disttree-$release from $ldc_name to $rdc_name},
        q{Distribution completed:},
        qq{Target Path = $rdc_name:/efs/dist/$mp_path/disttree-$release/common},
        q{Transfered 1 of 2 total files, 0 of 0 total bytes},
        q{Paths updated:},
        q{foo},
        qr{Elapsed \s+ time \s+ = \s+ \d+ \s+ secs}msx,
        qq{Successfully distributed $mp_path/disttree-$release/common to $rdc_name},
        qq{Successfully distributed the following installs to cell $rdc_name:},
        qq{$mp_path/disttree-$release/common},
        qq{Successfully distributed all installs to $rdc_name},
        q{Successfully distributed release to all cells},
    );

    return @output;

}

sub dist_release_failure {

    my $release = shift;

    my $datere = q{\[\d{4}-\d{2}-\d{2}\s+\d{2}:\d{2}:\d{2}\]};

    return (
        q{Dev environment diststate:},
        qr{Install \s+ $ldc_shortname \s+ $rdc_shortname}xms,
        qr{^=+}msx,
        qr{common \s+ P \s+ P}xms,
        qr{^-+}msx,
        q{Legend:},
        q{D: Item is disted and up to date},
        q{P: Item is disted but needs update},
        q{X: Item has not been disted},
        qq{Validating the checkpoint state for release '$mp_path/disttree-$release'.},
        qr{The\s+last\s+checktime\s+for\s+this\s+install\s+was\s+$datere}msx,
        q{The following paths have been modified since then:},
        qr{$datere\s+=>\s+/efs/dev/$mp_path/disttree-$release/install/common}msx,
        qq{Install '$mp_path/disttree-$release/common' has been modified since the last checktime.},
        q{The following installs have invalid checktime states:},
        qq{$mp_path/disttree-$release/common},
        qr{Error\s+running\s+EFS::CLI::Release->dist\s+via\s+EFS::RPC\s+on}msx,
    );

}

sub dist_releasealias_output {

    my ( $from, $to ) = @_;

    my $releasealias = qq{disttree-$from};
    my $target       = qq{disttree-$to};

    return (
        qq{Releasealias target: $target},
        q{Dev environment diststate:},
        qr{Cell \s+ Target}xms,
        qr{^=+}msx,
        qr{$ldc_shortname \s+ ---}xms,
        qr{$rdc_shortname \s+ ---}xms,
        qr{^-+}msx,
        q{Legend:},
        q{NNN: Releasealias is disted but target is not up to date},
        q{---: Releasealias is missing from cell},
        qq{Created symlink $ldc_name:/efs/dist/$mp_path/$releasealias -> $target},
        qq{Distributed releasealias $mp_path/$releasealias -> $target to $ldc_name},
        qq{Created symlink $rdc_name:/efs/dist/$mp_path/$releasealias -> $target},
        qq{Distributed releasealias $mp_path/$releasealias -> $target to $rdc_name},
        qq{Successfully distributed releasealias $mp_path/$releasealias to all cells},
    );

}
