#!/usr/bin/env perl

use strict;
use warnings;
use English;

use Test::More tests => 79;

use lib q{test/lib};
use EFS::Test;

my $test = EFS::Test->new;
my $config = $test->get_config;

my $ldc_shortname = $config->get_value( q{LocalDevCell}, q{shortname} );

my $base_mpr_path = q{testmeta/testproj/testrel};
my $base_mpr      = q{testmeta testproj testrel};

my $dep_mpr_path = qq{$base_mpr_path-dependency};
my $dep_mpr      = qq{$base_mpr-dependency};

my @setup = (
    q{efs display metaproj testmeta},
    q{efs display project testmeta testproj},
    qq{efs create release $base_mpr},
    qq{efs create install $base_mpr common},
    qq{mkdir /efs/dev/$base_mpr_path/install/common/bin},
    qq{touch /efs/dev/$base_mpr_path/install/common/bin/foobar.txt},
    qq{efs create release $dep_mpr},
    qq{efs create install $dep_mpr common},
    qq{mkdir /efs/dev/$dep_mpr_path/install/common/bin},
    qq{touch /efs/dev/$dep_mpr_path/install/common/bin/foobar.txt},
);

foreach my $setup (@setup) {
    $test->efs_test_command(
        name    => qq{(setup) $setup},
        command => $setup,
        bail    => 1,
    );
}

$test->efs_test_command(
    name    => q{dev->dev (no-op)},
    command => qq{efs setstage release $base_mpr dev},
    expect  => q{Release stage is already set to dev},
);

$test->efs_test_command(
    name    => q{dev->deprecated (not allowed)},
    command => qq{efs setstage release $base_mpr deprecated},
    not     => 1,
    expect  => q{Unsupported stage transition dev -> deprecated},
);

$test->efs_test_command(
    name    => q{dev->qa},
    command => qq{efs setstage release $base_mpr qa},
    expect  => qq{Changed the stage of $base_mpr_path from dev to qa},
);

$test->efs_test_command(
    name    => q{qa->qa (no-op)},
    command => qq{efs setstage release $base_mpr qa},
    expect  => q{Release stage is already set to qa},
);

$test->efs_test_command(
    name    => q{qa->deprecated (not allowed)},
    command => qq{efs setstage release $base_mpr deprecated},
    not     => 1,
    expect  => q{Unsupported stage transition qa -> deprecated},
);

$test->efs_test_command(
    name    => q{qa->rejected},
    command => qq{efs setstage release $base_mpr rejected},
    expect  => qq{Changed the stage of $base_mpr_path from qa to rejected},
);

$test->efs_test_command(
    name    => q{rejected->rejected (no-op)},
    command => qq{efs setstage release $base_mpr rejected},
    expect  => q{Release stage is already set to rejected},
);

$test->efs_test_command(
    name    => q{rejected->deprecated (not allowed)},
    command => qq{efs setstage release $base_mpr deprecated},
    not     => 1,
    expect  => q{Unsupported stage transition rejected -> deprecated},
);

$test->efs_test_command(
    name    => q{rejected->prod (not allowed)},
    command => qq{efs setstage release $base_mpr prod},
    not     => 1,
    expect  => q{Unsupported stage transition rejected -> prod},
);

$test->efs_test_command(
    name    => q{rejected->qa},
    command => qq{efs setstage release $base_mpr qa},
    expect  => qq{Changed the stage of $base_mpr_path from rejected to qa},
);

$test->efs_test_command(
    name    => q{qa->dev},
    command => qq{efs setstage release $base_mpr dev},
    expect  => qq{Changed the stage of $base_mpr_path from qa to dev},
);

$test->efs_test_command(
    name    => q{dev->rejected},
    command => qq{efs setstage release $base_mpr rejected},
    expect  => qq{Changed the stage of $base_mpr_path from dev to rejected},
);

$test->efs_test_command(
    name    => q{rejected->dev},
    command => qq{efs setstage release $base_mpr dev},
    expect  => qq{Changed the stage of $base_mpr_path from rejected to dev},
);

# NOTE: At this point, we've tested all the easy cases, that don't
# involve the transition to prod.

# [1] dev/qa -> prod requires a valid checkpoint

foreach my $stage (qw( dev qa )) {

    $test->efs_test_command(
        command => qq{efs setstage release $base_mpr $stage},
    );

    $test->efs_test_command(
        name    => qq{$stage->prod (missing checkpoint)},
        command => qq{efs setstage release $base_mpr prod},
        not     => 1,
        expect  => [
            qq{ Validating the checkpoint state for release '$base_mpr_path'. },
            qq{ Install '$base_mpr_path/common' has no checktime. },
            q{ The following installs have invalid checktime states: },
            qq{ $base_mpr_path/common },
        ],
    );

}

$test->efs_test_command(
    command => qq{efs checkpoint release $base_mpr},
);

# [2] dev/qa -> prod requires the release to have been disted

foreach my $stage (qw( dev qa )) {

    $test->efs_test_command(
        command => qq{efs setstage release $base_mpr $stage},
    );

    $test->efs_test_command(
        name    => qq{$stage->prod (missing dist)},
        command => qq{efs setstage release $base_mpr prod},
        not     => 1,
        expect  => [
            qq{ Validating the checkpoint state for release '$base_mpr_path'. },
            qq{ Validating the audit state for release '$base_mpr_path'. },
            q{ Unable to set the stage to prod, because the release has not been disted yet. },
        ],
    );

}

$test->efs_test_command(
    command => qq{efs dist release $base_mpr -cell $ldc_shortname},
    filter  => qr{to \s+ all \s+ cells}msx,
    expect  => q{Successfully distributed release to all cells},
);

# [3] If not disted to prod, and with no dependents, dev/qa -> prod
# -> dev/qa is allowed

foreach my $stage (qw( dev qa )) {

    $test->efs_test_command(
        command => qq{efs setstage release $base_mpr $stage},
    );

    $test->efs_test_command(
        name    => qq{$stage->prod},
        command => qq{efs setstage release $base_mpr prod},
        expect  => [
            qq{ Validating the checkpoint state for release '$base_mpr_path'. },
            qq{ Validating the audit state for release '$base_mpr_path'. },
            qq{Changed the stage of $base_mpr_path from $stage to prod},
        ],
    );

    $test->efs_test_command(
        name    => qq{prod->$stage},
        command => qq{efs setstage release $base_mpr $stage},
        expect  => qq{Changed the stage of $base_mpr_path from prod to $stage},
    );

}

# [4] With a non-prod dependency, dev/qa -> prod fails
# NOTE: We're attempting to change the stage of the dependent here

$test->efs_test_command(
    command => qq{efs define dependency $dep_mpr $base_mpr runtime},
);

$test->efs_test_command(
    command => qq{efs checkpoint release $dep_mpr},
);

$test->efs_test_command(
    command => qq{efs dist release $dep_mpr -cell $ldc_shortname},
);

foreach my $stage (qw( dev qa )) {

    $test->efs_test_command(
        command => qq{efs setstage release $dep_mpr $stage},
    );

    $test->efs_test_command(
        name    => qq{$stage->prod with dependent not yet prod},
        command => qq{efs setstage release $dep_mpr prod},
        not     => 1,
        expect  => [
            qq{ Validating the checkpoint state for release '$dep_mpr_path'. },
            qq{ Validating the audit state for release '$dep_mpr_path'. },
            q{The following dependencies are not prod or deprecated:},
            qq{$base_mpr_path},
        ],
    );

}

# [5] With a non-prod dependent, dev/qa -> prod works,
# but when dependent is then prod -> dev/qa fails

$test->efs_test_command(
    command => qq{efs undef dependency $dep_mpr $base_mpr runtime},
);

foreach my $stage (qw( dev qa )) {

    $test->efs_test_command(
        command => qq{efs setstage release $base_mpr $stage},
    );

    $test->efs_test_command(
        name    => qq{$stage->prod, should work with no dependents},
        command => qq{efs setstage release $base_mpr prod},
        expect  => [
            qq{ Validating the checkpoint state for release '$base_mpr_path'. },
            qq{ Validating the audit state for release '$base_mpr_path'. },
            qq{ Changed the stage of $base_mpr_path from $stage to prod },
        ],
    );

    $test->efs_test_command(
        command => qq{efs define dependency $dep_mpr $base_mpr runtime},
    );

    $test->efs_test_command(
        command => qq{efs setstage release $dep_mpr prod},
    );

    $test->efs_test_command(
        name    => qq{prod->$stage, fails because the dependent is prod},
        command => qq{efs setstage release $base_mpr $stage},
        not     => 1,
        expect  => [
            qq{ Release '$base_mpr_path' has prod/deprecated dependents, },
            q{ and therefore can NOT be reverted to dev or qa. },
        ],
    );

    $test->efs_test_command(
        command => qq{efs undef dependency $dep_mpr $base_mpr runtime},
    );

    $test->efs_test_command(
        command => qq{efs setstage release $dep_mpr $stage},
    );

}

# NOTE: We've now tested ALL of the scenarios where we allow prod -> dev/qa

# This section tests all the simple cases starting from prod/deprecated

$test->efs_test_command(
    command => qq{efs setstage release $base_mpr prod},
);

$test->efs_test_command(
    name    => q{prod->prod (no-op)},
    command => qq{efs setstage release $base_mpr prod},
    expect  => q{Release stage is already set to prod},
);

$test->efs_test_command(
    name    => q{prod->rejected (not allowed)},
    command => qq{efs setstage release $base_mpr rejected},
    not     => 1,
    expect  => q{Unsupported stage transition prod -> rejected},
);

$test->efs_test_command(
    name    => q{prod->deprecated},
    command => qq{efs setstage release $base_mpr deprecated},
    expect  => qq{Changed the stage of $base_mpr_path from prod to deprecated},
);

$test->efs_test_command(
    name    => q{deprecated->deprecated (no-op)},
    command => qq{efs setstage release $base_mpr deprecated},
    expect  => q{Release stage is already set to deprecated},
);

foreach my $stage (qw( dev qa rejected )) {

    $test->efs_test_command(
        name    => qq{deprecated->$stage (not allowed)},
        command => qq{efs setstage release $base_mpr $stage},
        not     => 1,
        expect  => qq{Unsupported stage transition deprecated -> $stage},
    );

}

$test->efs_test_command(
    name    => q{deprecated->prod},
    command => qq{efs setstage release $base_mpr prod},
    expect  => qq{Changed the stage of $base_mpr_path from deprecated to prod},
);

# Last thing to test: if disted to prod, prod->dev/qa is not allowed

$test->efs_test_command(
    command => qq{efs dist release $base_mpr --celltype prod --global},
);

foreach my $stage (qw( dev qa )) {

    $test->efs_test_command(
        name    => qq{prod->$stage (not allowed if disted to prod)},
        command => qq{efs setstage release $base_mpr $stage},
        not     => 1,
        expect  => [
            qq{Release '$base_mpr_path' has been disted to one or more},
            q{prod cells and can NOT be reverted to dev or qa.},
        ],
    );

}

my @teardown = (
    qq{efs setstage release $dep_mpr rejected},
    qq{efs destroy release $dep_mpr},
    qq{efs purge release $dep_mpr -now},
    qq{efs setstage release $base_mpr deprecated},
    qq{efs destroy release $base_mpr},
    qq{efs purge release $base_mpr -now},
);

foreach my $teardown (@teardown) {
    $test->efs_test_command(
        run_as  => q{efsops},
        name    => qq{(teardown) $teardown},
        command => $teardown,
    );
}

exit 0;
