#!/usr/bin/env perl

use strict;
use warnings;
use English;

use Test::More tests => 51;
use List::MoreUtils qw( uniq );

use lib q{test/lib};
use EFS::Test;

my $test = EFS::Test->new;
my $config = $test->get_config;

my $username  = $test->get_primary_username;
my $groupname = $test->get_primary_groupname;

my $root_domain = $config->get_value( q{Global}, q{root_domain} );

my $ldc_name      = $config->get_value( q{LocalDevCell}, q{name} );
my $ldc_shortname = $config->get_value( q{LocalDevCell}, q{shortname} );

my $base_mpr_path = q{testmeta/testproj/testrel};
my $base_mpr      = q{testmeta testproj testrel};

my $dep_mpr_path = q{testmeta/testprojdep/testrel};
my $dep_mpr      = q{testmeta testprojdep testrel};

my $from_address = qq{EFS_autonotify\@$root_domain};
my $owner        = qq{owner\@$root_domain};
my $owner2       = qq{owner2\@$root_domain};
my $notify       = qq{notify\@$root_domain};
my $notify2      = qq{notify2\@$root_domain};

my @setup = (
    q{efs display metaproj testmeta},
    qq{efs create project testmeta testproj $ldc_name},
    qq{efs create release $base_mpr},
    qq{efs create install $base_mpr common},
    qq{mkdir -p /efs/dev/$base_mpr_path/install/common/bin},
    qq{touch /efs/dev/$base_mpr_path/install/common/bin/foobar.txt},
    qq{efs create project testmeta testprojdep $ldc_name},
    qq{efs create release $dep_mpr},
    qq{efs create install $dep_mpr common},
    qq{mkdir -p /efs/dev/$dep_mpr_path/install/common/bin},
    qq{touch /efs/dev/$dep_mpr_path/install/common/bin/foobar.txt},
);

foreach my $setup (@setup) {
    $test->efs_test_command(
        name    => qq{(setup) $setup},
        command => $setup,
        bail    => 1,
    );
}

$test->efs_test_command(
    run_as  => q{efsops},
    command => qq{efs define parameter notification_from_address '$from_address'},
);

$test->efs_test_command(
    run_as  => q{efsops},
    command => qq{efs define parameter notification_email_domain '$root_domain'},
);

$test->efs_test_command(
    run_as  => q{efsops},
    command => qq{efs undef userrights testmeta admin $username},
);

$test->efs_test_command(
    name    => q{Create owner-list for testmeta},
    command => qq{efs set_owner_list metaproj testmeta $owner},
    not     => 1,
    expect  => [
        q{Access denied.},
        q{The 'admin' role is required to set_owner_list metaproj},
        q{Exit status from efs server not received.  Assuming failure...},
    ],
);

$test->efs_test_command(
    run_as  => q{efsops},
    command => qq{efs define userrights testmeta admin $username},
);

$test->efs_test_command(
    name    => q{Create owner-list for testmeta},
    command => qq{efs set_owner_list metaproj testmeta $owner},
);

$test->efs_test_command(
    name    => q{Check owner-list for testmeta},
    command => q{efs get_owner_list metaproj testmeta},
    expect  => qq{$owner},
);

$test->efs_test_command(
    name    => q{Make sure owner bubbles down},
    command => q{efs get_owner_list project testmeta testproj},
    expect  => qq{$owner (Inherited from testmeta)},
);

$test->efs_test_command(
    name    => q{Create duplicate owner-list for testproj},
    command => qq{efs set_owner_list project testmeta testproj $owner}
);

$test->efs_test_command(
    name    => q{Make sure owner is now set normally},
    command => q{efs get_owner_list project testmeta testproj},
    expect  => qq{$owner},
);

$test->efs_test_command(
    name    => q{Create notify-list for testmeta},
    command => qq{efs set_notify_list metaproj testmeta $notify}
);

$test->efs_test_command(
    name    => q{Check notify-list for testmeta},
    command => q{efs get_notify_list metaproj testmeta},
    expect  => qq{$notify},
);

$test->efs_test_command(
    name    => q{Make sure notify bubbles down},
    command => q{efs get_notify_list project testmeta testproj},
    expect  => qq{$notify (Inherited from testmeta)},
);

$test->efs_test_command(
    name    => q{Create non-duplicate notify-list for testproj},
    command => qq{efs set_notify_list project testmeta testproj $notify2}
);

$test->efs_test_command(
    name    => q{Check notify-list for testproj},
    command => q{efs get_notify_list project testmeta testproj},
    expect  => qq{$notify2},
);

$test->efs_test_command(
    name    => q{checkpoint},
    command => qq{efs checkpoint release $base_mpr},
);

$test->efs_test_command(
    name    => q{dist},
    command => qq{efs dist release $base_mpr -cell $ldc_shortname},
    filter  => qr{to \s+ all \s+ cells}msx,
    expect  => q{Successfully distributed release to all cells},
);

# Dev, so no mail should be generated

$test->efs_test_command(
    name    => q{unlock with dev release, no mail generated},
    command => qq{efs unlock release $base_mpr},
    filter  => qr{TEST}msx,
    expect  => q{},
);

$test->efs_test_command(
    name    => q{checkpoint},
    command => qq{efs checkpoint release $base_mpr},
);

$test->efs_test_command(
    name    => q{setstage prod},
    command => qq{efs setstage release $base_mpr prod},
);

my $to = join(
    q{, },
    uniq( sort $notify2, $notify, $owner, qq{$username\@$root_domain} )
);

$test->efs_test_command(
    name    => q{unlock with prod release, mail generated},
    command => qq{efs unlock release $base_mpr},
    filter  => qr{^TEST:\s+.*(From:|To:|Cc:|Reply-To:|Priority|Subject:|by\s+the\s+user)}msx,
    expect  => [
        qq{TEST: From: $from_address},
        qq{TEST: To: $to},
        qq{TEST: Cc: $from_address},
        qq{TEST: Reply-To: $owner},
        q{TEST: Priority: Highest},
        qq{TEST: Subject: EFS Notification: $base_mpr_path UNLOCKED},
        qr{^TEST:\s+unlocked\s+by\s+the\s+user\s+"\S+}msx,
    ],
);

$test->efs_test_command(
    name    => q{Relock release},
    command => qq{efs lock release $base_mpr}
);

## Check dependency

$test->efs_test_command(
    name    => q{Set a dependency on the release},
    command => qq{efs define dependency testmeta testprojdep testrel $base_mpr runtime}
);

$test->efs_test_command(
    name    => q{Check that dependent gets owner from metaproj},
    command => q{efs get_owner_list project testmeta testprojdep},
    expect  => qq{$owner (Inherited from testmeta)},
);

$test->efs_test_command(
    name    => q{Set different owner for dependent},
    command => qq{efs set_owner_list project testmeta testprojdep $owner2}
);

$test->efs_test_command(
    name    => q{Check that dependent's owner is set},
    command => q{efs get_owner_list project testmeta testprojdep},
    expect  => qq{$owner2},
);

$to = join(
    q{, },
    uniq( sort $notify2, $notify, $owner, $owner2, qq{$username\@$root_domain} )
);

$test->efs_test_command(
    name    => q{unlock prod release with dependent, check mail},
    command => qq{efs unlock release $base_mpr},
    filter  => qr{^TEST:\s+.*(From:|To:|Cc:|Reply-To:|Subject:|by\s+the\s+user|direct\s+dependent)}msx,
    expect  => [
        qq{TEST: From: $from_address},
        qq{TEST: To: $to},
        qq{TEST: Cc: $from_address},
        qq{TEST: Reply-To: $owner},
        qq{TEST: Subject: EFS Notification: $base_mpr_path UNLOCKED},
        qr{^TEST:\s+unlocked\s+by\s+the\s+user\s+"\S+}msx,
        qr{direct\s+dependents\s+of\s+the\s+above\s+release}msx,
        qr{$dep_mpr_path\s+[(]direct\s+dependent[)]}msx,
    ],
);

$test->efs_test_command(
    name    => q{Clear notify-list for testproj},
    command => q{efs clear_notify_list project testmeta testproj},
);

$test->efs_test_command(
    name    => q{Make sure notify bubbles down again},
    command => q{efs get_notify_list project testmeta testproj},
    expect  => qq{$notify (Inherited from testmeta)},
);

$test->efs_test_command(
    name    => q{Remove owner_list from testproj},
    command => q{efs clear_owner_list project testmeta testproj}
);

$test->efs_test_command(
    name    => q{Ensure we can't remove owner_list from testmeta},
    command => q{efs clear_owner_list metaproj testmeta},
    not     => 1,
);

# Cleanup

my @teardown = (
    qq{efs setstage release $dep_mpr rejected},
    qq{efs destroy release $dep_mpr},
    qq{efs purge release $dep_mpr -now},
    q{efs destroy project testmeta testprojdep},
    qq{efs setstage release $base_mpr deprecated},
    qq{efs destroy release $base_mpr},
    qq{efs purge release $base_mpr -now},
    q{efs undef parameter notification_from_address},
    q{efs undef parameter notification_email_domain},
);

foreach my $teardown (@teardown) {
    $test->efs_test_command(
        run_as  => q{efsops},
        name    => qq{(teardown) $teardown},
        command => $teardown,
    );
}

exit 0;

