#!/usr/bin/env perl

## no critic: RequireExtendedFormatting

use strict;
use warnings;
use English;

use File::Slurp;

use lib q{test/lib};
use EFS::Test;
use EFS::Test::Config;

use blib;
use EFS::Util;

use Test::More tests => 56;

my $test   = EFS::Test->new;
my $config = EFS::Test::Config->new;

my $upload_server   = $config->get_value( qw( Packaging upload_server ) );
my $upload_username = $config->get_value( qw( Packaging upload_username ) );
my $upload_sshkey   = $config->get_value( qw( Packaging upload_sshkey ) );
my $upload_root     = $config->get_value( qw( Packaging upload_root ) );

my $download_url    = $config->get_value( qw( Packaging download_url ) );
my ($download_dir)  = ( $download_url =~ m{^file://localhost(.*)}msx );

purge_download_dirs(
    stage    => q{setup, phase one},
    projects => [qw( upload download )],
);

my @setup = (
    q{efs create metaproj testpkgs efsops efsops},
);

push @setup, (
    qq{efs create project testpkgs upload},
    qq{efs create release testpkgs upload testrel},
);

foreach my $install ( qw( common x86-64.rhel.5 x86-32.rhel.5 ) ) {

    push @setup, qq{efs create install testpkgs upload testrel $install};

    my $install_dir = qq{/efs/dev/testpkgs/upload/testrel/install/};
    $install_dir .= $install eq q{common} ? $install : qq{.exec/$install};

    push @setup, qq{touch $install_dir/notempty};

    my $build_dir = qq{/efs/dev/testpkgs/upload/testrel/build/$install};
    push @setup, (
        qq{mkdir $build_dir},
        qq{touch $build_dir/notempty},
    );

}

my $src_dir = qq{/efs/dev/testpkgs/upload/testrel/src};
push @setup, qq{touch $src_dir/notempty};

push @setup, qq{efs checkpoint release testpkgs upload testrel};

my %attrs = (
    upload_server   => $upload_server,
    upload_username => $upload_username,
    upload_sshkey   => $upload_sshkey,
    upload_root     => $upload_root,
    download_url    => $download_url,
);

foreach my $key ( sort keys %attrs ) {
    push @setup, qq{efs setattr metaproj testpkgs $key $attrs{$key}};
}

foreach my $command ( @setup ) {
    $test->efs_test_command(
        run_as  => q{efsops},
        name    => qq{(setup, phase one) $command},
        command => $command,
        bail    => 1,
    );
}

# We need to force the addition of the host key, so that subsequent
# commands have predictable output.  NOTE: We do this in all 3 of
# these packaging tests so that they can be run independently.
$test->efs_test_command(
    run_as  => q{root},
    name    => q{(setup, phase one) Initializing ssh host key},
    command => qq{ssh -o StrictHostKeyChecking=no -i $upload_sshkey -l $upload_username $upload_server true},
    bail    => 1,
);

# End of setup

my @attrs = split m{\s+}msx, q{install install,src install,build,src all};

foreach my $attr ( @attrs ) {

    if ( $attr ne q{install} ) {
        $test->efs_test_command(
            run_as  => q{efsops},
            command => qq{efs setattr metaproj testpkgs upload_content $attr},
        );
    }

    $test->efs_test_command(
        run_as  => q{efsops},
        command => qq{efs upload release testpkgs upload testrel --packageonly --force},
        expect  => [ package_release_output( $attr ) ],
    );

}

$test->efs_test_command(
    run_as  => q{efsops},
    command => qq{efs upload release testpkgs upload testrel},
);

# End of upload tests

convert_upload_to_download();

# Start of download tests

foreach my $attr ( @attrs ) {

    if ( $attr ne q{install} ) {
        $test->efs_test_command(
            run_as  => q{efsops},
            command => qq{efs setattr metaproj testpkgs download_content $attr},
        );
        $test->efs_test_command(
            run_as  => q{efsops},
            command => qq{efs unlock release testpkgs download testrel},
        );
    }

    $test->efs_test_command(
        run_as  => q{efsops},
        command => qq{efs download release testpkgs download testrel},
        filter  => qr{Downloading $download_url}ms,
        expect  => [ download_release_output( $attr ) ],
    );

}

# Start of teardown

my @teardown = ();

foreach my $project ( qw( upload download ) ) {
    push @teardown, (
        qq{efs setstage release testpkgs $project testrel rejected},
        qq{efs destroy release testpkgs $project testrel},
        qq{efs purge release testpkgs $project testrel --now},
        qq{efs destroy project testpkgs $project},
    );
}

push @teardown, q{efs destroy metaproj testpkgs};

foreach my $command ( @teardown ) {
    $test->efs_test_command(
        run_as  => q{efsops},
        name    => qq{(teardown) $command},
        command => $command,
    );
}

purge_download_dirs(
    stage    => q{teardown},
    projects => [qw( upload download )],
);

exit 0;

sub purge_download_dirs {

    my %args = @_;

    foreach my $project ( @{ $args{projects} } ) {
        my $dir = qq{$download_dir/testpkgs/$project};
        my $command = qq{su $upload_username -c 'rm -rf $dir'};
        $test->efs_test_command(
            run_as  => q{root},
            name    => qq{($args{stage}) $command},
            command => $command,
            bail    => 1,
        );
    }

    return 1;

}

sub package_release_output {

    my $attr = shift;

    my %content = map { $_ => 1 } split m{,}msx, $attr;

    if ( $content{all} ) {
        $content{install}++;
        $content{build}++;
        $content{src}++;
    }

    my $directory = qq{/efs/dev/testpkgs/upload/testrel/build/upload};

    my @output = (
        qq{Validating the checkpoint state for release 'testpkgs/upload/testrel'.},
        qr{(Removed|No such) directory $directory}ms,
        qq{Created directory $directory},
        qq{Created directory $directory/package},
    );

    if ( $content{install} ) {
        push @output, (
            qq{Created directory $directory/package/install},
            qq{Creating $directory/package/install/common.tar.bz2},
            qq{Creating $directory/package/install/x86-32.rhel.5.tar.bz2},
            qq{Creating $directory/package/install/x86-64.rhel.5.tar.bz2},
        );
    }

    if ( $content{build} ) {
        push @output, (
            qq{Created directory $directory/package/build},
            qq{Creating $directory/package/build/x86-32.rhel.5.tar.bz2},
            qq{Creating $directory/package/build/x86-64.rhel.5.tar.bz2},
            qq{Creating $directory/package/build/common.tar.bz2},
        );
    }

    if ( $content{src} ) {
        push @output, qq{Creating $directory/package/src.tar.bz2};
    }

    push @output, (
        qq{Packaged testpkgs/upload/testrel},
    );

    return @output;

}

sub download_release_output {

    my $attr = shift;

    my %content = map { $_ => 1 } split m{,}msx, $attr;

    if ( $content{all} ) {
        $content{install}++;
        $content{build}++;
        $content{src}++;
    }

    my @output = (
        qq{Downloading $download_url/testpkgs/download/testrel/metadata.conf},
    );

    if ( $content{src} ) {
        push @output, (
            qq{Downloading $download_url/testpkgs/download/testrel/src.tar.bz2},
        );
    }

    if ( $content{install} ) {
        push @output, (
            qq{Downloading $download_url/testpkgs/download/testrel/install/common.tar.bz2},
            qq{Downloading $download_url/testpkgs/download/testrel/install/x86-32.rhel.5.tar.bz2},
            qq{Downloading $download_url/testpkgs/download/testrel/install/x86-64.rhel.5.tar.bz2},
        );
    }

    if ( $content{build} ) {
        push @output, (
            qq{Downloading $download_url/testpkgs/download/testrel/build/x86-32.rhel.5.tar.bz2},
            qq{Downloading $download_url/testpkgs/download/testrel/build/x86-64.rhel.5.tar.bz2},
            qq{Downloading $download_url/testpkgs/download/testrel/build/common.tar.bz2},
        );
    }

    return @output;

}

sub convert_upload_to_download {

    my $rsync = qq{rsync -az $download_dir/testpkgs/upload/ $download_dir/testpkgs/download};
    my $command = qq{su $upload_username -c '$rsync'};

    # XXX: Workaround -- on debian, we're not seeing this directory
    # immediately, most likely due to NFS client side caching.
    EFS::Util->WaitDir( directory => qq{$download_dir/testpkgs/upload} );

    $test->efs_test_command(
        run_as  => q{root},
        name    => qq{(setup, phase two) $command},
        command => $command,
        bail    => 1,
    );

    my $file = qq{$download_dir/testpkgs/download/testrel/metadata.conf};
    my $temp = qq{/var/tmp/testpkgs-download-testrel-metadata.conf};

    if ( -f $temp ) {
        unlink $temp || die qq{Unable to unlink $temp: $ERRNO\n};
    }

    my $metadata = read_file( $file );
    $metadata =~ s{upload}{download}gmsx;
    write_file( $temp, $metadata );

    $test->efs_test_command(
        run_as  => q{root},
        name    => qq{(setup, phase two) edit $file},
        command => qq{su $upload_username -c 'cp $temp $file'},
    );

    unlink $temp || die qq{Unable to unlink $temp: $ERRNO\n};

}
