#!/usr/bin/env perl

use strict;
use warnings;
use English;

use File::Basename qw( dirname );

use Test::More;

use lib q{test/lib};
use EFS::Test;

my $test = EFS::Test->new( as_user => q{efsops} );
my $config = $test->get_config;

my $username  = $test->get_primary_username;
my $groupname = $test->get_primary_groupname;

my $root_domain = $config->get_value( q{Global}, q{root_domain} );
my $ownerlist   = q{efsops@} . $root_domain;

my $ldc_name       = $config->get_value( q{LocalDevCell}, q{name} );
my $ldc_shortname  = $config->get_value( q{LocalDevCell}, q{shortname} );
my $ldc_fileserver = $config->get_value( q{LocalDevCell}, q{fileserver} );
my $ldc_devshare   = $config->get_value( q{LocalDevCell}, q{devshare} );
my $ldc_homeshare  = $config->get_value( q{LocalDevCell}, q{homeshare} );
my $ldc_distshare  = $config->get_value( q{LocalDevCell}, q{distshare} );

my $rdc_name       = $config->get_value( q{RemoteDevCell}, q{name} );
my $rdc_shortname  = $config->get_value( q{RemoteDevCell}, q{shortname} );
my $rdc_fileserver = $config->get_value( q{RemoteDevCell}, q{fileserver} );
my $rdc_devshare   = $config->get_value( q{RemoteDevCell}, q{devshare} );
my $rdc_homeshare  = $config->get_value( q{RemoteDevCell}, q{homeshare} );
my $rdc_distshare  = $config->get_value( q{RemoteDevCell}, q{distshare} );

my $lpc_name       = $config->get_value( q{LocalProdCell}, q{name} );
my $lpc_shortname  = $config->get_value( q{LocalProdCell}, q{shortname} );
my $lpc_fileserver = $config->get_value( q{LocalProdCell}, q{fileserver} );
my $lpc_distshare  = $config->get_value( q{LocalProdCell}, q{distshare} );

my @setup = (
    qq{efs create metaproj testexec efsops efsops $ownerlist},
    qq{efs create project testexec testexec $ldc_name},
);

foreach my $release ( qw( platform instance none
                          backwards emulates both ) ) {
    push @setup, qq{efs create release testexec testexec $release};
}

push @setup, (
    q{efs setattr release testexec testexec backwards } .
        q{disable_exec_backwards true},
    q{efs setattr release testexec testexec emulates } .
        q{disable_exec_emulates true},
    q{efs setattr release testexec testexec both } .
        q{disable_exec_emulates true},
    q{efs setattr release testexec testexec both } .
        q{disable_exec_backwards true},
);

foreach my $setup (@setup) {
    $test->efs_test_command(
        name    => "(setup) $setup",
        command => $setup,
        bail    => 1,
    );
}

my $pl32_1 = {
    name    => q{x86-32.someos.1}, # -> oldname
    newname => q{x86-32.newos.1}, # -> name
};

my $pl64_1 = {
    name     => q{x86-64.someos.1},
    emulates => $pl32_1->{name},
};

my $pl32_2 = {
    name      => q{x86-32.someos.2},
    backwards => $pl32_1->{name},
};

my $pl64_2 = {
    name      => q{x86-64.someos.2},
    backwards => $pl64_1->{name},
    emulates  => $pl32_2->{name},
};

my $pa64_2 = {
    name     => q{x86-64.aliasos.2},
    platform => $pl64_2->{name},
};

my $in32_1 = {
    name     => q{amd32.someos.1}, # -> oldname
    platform => $pl32_1->{name},
    newname  => q{amd32.newos.1}, # -> name
};

my $in64_1 = {
    name     => q{amd64.someos.1},
    platform => $pl64_1->{name},
    emulates => $in32_1->{name},
};

my $in32_2 = {
    name      => q{amd32.someos.2},
    platform  => $pl32_2->{name},
    backwards => $in32_1->{name},
};

my $in64_2 = {
    name      => q{amd64.someos.2},
    platform  => $pl64_2->{name},
    backwards => $in64_1->{name},
    emulates  => $in32_2->{name},
};

my $ia64_2 = {
    name             => q{amd64.aliasos.2},
    platforminstance => $in64_2->{name},
};

my $link_exists = q{};

# Test platform, instance and alias definition basics

$test->efs_test_command(
    name    => q{Define the base platform},
    command => qq{efs define platform $pl32_1->{name}},
    expect  => qq{Defined platform $pl32_1->{name}},
);

$test->efs_test_command(
    name    => q{Unable to define platform when backwards doesn't exist},
    command => qq{efs define platform $pl64_1->{name} nosuchplatform},
    not     => 1,
    expect  => q{No such platform 'nosuchplatform'},
);

$test->efs_test_command(
    name    => q{Unable to define platform when emulates doesn't exist},
    command => qq{efs define platform $pl64_1->{name} --emulates nosuchplatform},
    not     => 1,
    expect  => q{No such platform 'nosuchplatform'},
);

$test->efs_test_command(
    name    => q{Unable to define platform when backwards eq name},
    command => qq{efs define platform $pl64_1->{name} $pl64_1->{name}},
    not     => 1,
    expect  => [
        qq{Invalid backwards platform '$pl64_1->{name}'},
        qq{Matches platform name '$pl64_1->{name}'},
    ],
);

$test->efs_test_command(
    name    => q{Unable to define platform when emulates eq name},
    command => qq{efs define platform $pl64_1->{name} --emulates $pl64_1->{name}},
    not     => 1,
    expect  => [
        qq{Invalid emulates platform '$pl64_1->{name}'},
        qq{Matches platform name '$pl64_1->{name}'},
    ],
);

$test->efs_test_command(
    name    => q{Define a platform without the emulates...},
    command => qq{efs define platform $pl64_1->{name}},
    expect  => qq{Defined platform $pl64_1->{name}},
);

$test->efs_test_command(
    name    => q{...and then update it},
    command => qq{efs define platform $pl64_1->{name} --emulates $pl64_1->{emulates}},
    expect  => [
        qq{Platform $pl64_1->{name} exists},
        qq{Updated platform, set emulates => $pl64_1->{emulates}},
    ],
);

$test->efs_test_command(
    name    => q{Define a platform without the backwards...},
    command => qq{efs define platform $pl32_2->{name}},
    expect  => qq{Defined platform $pl32_2->{name}},
);

$test->efs_test_command(
    name    => q{...and then update it},
    command => qq{efs define platform $pl32_2->{name} $pl32_2->{backwards}},
    expect  => [
        qq{Platform $pl32_2->{name} exists},
        qq{Updated platform, set backwards => $pl32_2->{backwards}},
    ],
);

$test->efs_test_command(
    name    => q{Define the last platform we need for these tests},
    command => qq{efs define platform $pl64_2->{name} $pl64_2->{backwards} $pl64_2->{emulates}},
);

# That completes all the basic platform definition tests.
# Now we do instances

$test->efs_test_command(
    name    => q{Unable to define instance using an existing platform name},
    command => qq{efs define platforminstance $pl32_1->{name} $pl32_1->{name}},
    not     => 1,
    expect  => qq{Invalid argument: $pl32_1->{name} is already a platform},
);

$test->efs_test_command(
    name    => q{Unable to define instance when the platform doesn't exist},
    command => qq{efs define platforminstance $in32_1->{name} nosuchplatform},
    not     => 1,
    expect  => q{No such platform 'nosuchplatform'},
);

$test->efs_test_command(
    name    => q{Unable to define instance when backwards doesn't exist},
    command => qq{efs define platforminstance $in32_1->{name} $in32_1->{platform} nosuchinstance},
    not     => 1,
    expect  => q{No such platform instance 'nosuchinstance'},
);

$test->efs_test_command(
    name    => q{Unable to define instance when emulates doesn't exist},
    command => qq{efs define platforminstance $in32_1->{name} $in32_1->{platform} --emulates nosuchinstance},
    not     => 1,
    expect  => q{No such platform instance 'nosuchinstance'},
);

$test->efs_test_command(
    name    => q{Unable to define instance when backwards eq name},
    command => qq{efs define platforminstance $in32_1->{name} $in32_1->{platform} $in32_1->{name}},
    not     => 1,
    expect  => [
        qq{Invalid backwards platform '$in32_1->{name}'},
        qq{Matches platform instance name '$in32_1->{name}'},
    ],
);

$test->efs_test_command(
    name    => q{Unable to define instance when emulates eq name},
    command => qq{efs define platforminstance $in32_1->{name} $in32_1->{platform} --emulates $in32_1->{name}},
    not     => 1,
    expect  => [
        qq{Invalid emulates platform '$in32_1->{name}'},
        qq{Matches platform instance name '$in32_1->{name}'},
    ],
);

$test->efs_test_command(
    name    => q{Define the base instance},
    command => qq{efs define platforminstance $in32_1->{name} $in32_1->{platform} },
    expect  => qq{Defined platform instance $in32_1->{name}},
);

$test->efs_test_command(
    name    => q{Define an instance without the emulates...},
    command => qq{efs define platforminstance $in64_1->{name} $in64_1->{platform} },
    expect  => qq{Defined platform instance $in64_1->{name}},
);

$test->efs_test_command(
    name    => q{...and then update it},
    command => qq{efs define platforminstance $in64_1->{name} $in64_1->{platform} --emulates $in64_1->{emulates}},
    expect  => [
        qq{Platform instance $in64_1->{name} exists},
        qq{Updated platform instance, set emulates => $in64_1->{emulates}},
    ],
);

$test->efs_test_command(
    name    => q{Define an instance without the backwards...},
    command => qq{efs define platforminstance $in32_2->{name} $in32_2->{platform} },
    expect  => qq{Defined platform instance $in32_2->{name}},
);

$test->efs_test_command(
    name    => q{...and then update it},
    command => qq{efs define platforminstance $in32_2->{name} $in32_2->{platform} $in32_2->{backwards}},
    expect  => [
        qq{Platform instance $in32_2->{name} exists},
        qq{Updated platform instance, set backwards => $in32_2->{backwards}},
    ],
);

$test->efs_test_command(
    name    => q{Define the last instance we need for these tests},
    command => qq{efs define platforminstance $in64_2->{name} $in64_2->{platform} $in64_2->{backwards} $in64_2->{emulates}},
);

# That completes all the basic instance definition tests
# Now we do aliases

$test->efs_test_command(
    name    => q{Unable to define platform alias using existing platform name},
    command => qq{efs define platformalias $pl64_2->{name} $pa64_2->{platform}},
    not     => 1,
    expect  => qq{Invalid argument: $pl64_2->{name} is already a platform},
);

$test->efs_test_command(
    name    => q{Unable to define platform alias when platform doesn't exist},
    command => qq{efs define platformalias $pa64_2->{name} nosuchplatform},
    not     => 1,
    expect  => q{No such platform 'nosuchplatform'},
);

$test->efs_test_command(
    name    => q{Define the platform alias},
    command => qq{efs define platformalias $pa64_2->{name} $pa64_2->{platform}},
    expect  => qq{Defined $pa64_2->{name} as an alias for $pa64_2->{platform}},
);

$test->efs_test_command(
    name    => q{Unable to define instance alias using existing instance name},
    command => qq{efs define platforminstancealias $in32_1->{name} $ia64_2->{platforminstance}},
    not     => 1,
    expect  =>
        qq{Invalid argument: $in32_1->{name} is already a platforminstance},
);

$test->efs_test_command(
    name    => q{Unable to define instance alias when instance doesn't exist},
    command => qq{efs define platforminstancealias $ia64_2->{name} nosuchinstance},
    not     => 1,
    expect  => q{No such platform instance 'nosuchinstance'},
);

$test->efs_test_command(
    name    => q{Define the instance alias},
    command => qq{efs define platforminstancealias $ia64_2->{name} $ia64_2->{platforminstance}},
    expect  => qq{Defined $ia64_2->{name} as an alias for $ia64_2->{platforminstance}},
);

# That completes ALL of the basic definition tests

# Test the init fileqtree command, make sure the new instances, and
# aliases get added

$test->efs_test_command(
    command => qq{efs init fileqtree $ldc_name $ldc_fileserver $ldc_devshare efs_qtree},
    expect  => [
        qq{Created symlink $ldc_name:/.efs/$ia64_2->{name} -> $in64_2->{name}},
        qq{Created symlink $ldc_name:/.efs/$pa64_2->{name} -> $pl64_2->{name}},
        qq{Created directory $ldc_name:/.efs/$in32_1->{name}},
        qq{Created directory $ldc_name:/.efs/$in32_1->{name}/dev},
        qq{Created directory $ldc_name:/.efs/$in32_2->{name}},
        qq{Created directory $ldc_name:/.efs/$in32_2->{name}/dev},
        qq{Created directory $ldc_name:/.efs/$in64_1->{name}},
        qq{Created directory $ldc_name:/.efs/$in64_1->{name}/dev},
        qq{Created directory $ldc_name:/.efs/$in64_2->{name}},
        qq{Created directory $ldc_name:/.efs/$in64_2->{name}/dev},
        qq{Created directory $ldc_name:/.efs/$pl32_1->{name}},
        qq{Created directory $ldc_name:/.efs/$pl32_1->{name}/dev},
        qq{Created directory $ldc_name:/.efs/$pl32_2->{name}},
        qq{Created directory $ldc_name:/.efs/$pl32_2->{name}/dev},
        qq{Created directory $ldc_name:/.efs/$pl64_1->{name}},
        qq{Created directory $ldc_name:/.efs/$pl64_1->{name}/dev},
        qq{Created directory $ldc_name:/.efs/$pl64_2->{name}},
        qq{Created directory $ldc_name:/.efs/$pl64_2->{name}/dev},
        qq{Initialized fileqtree $ldc_fileserver:$ldc_devshare/efs_qtree},
        q{Initialized all fileqtrees},
    ],
);

# NOTE: The predictability of the output depends on the hostnames, due
# to how the fileqtrees are sorted.

$test->efs_test_command(
    command => qq{efs init fileqtree $ldc_name},
    filter  => qr{^ Initial}msx,
    expect  => [
        qq{Initialized fileqtree $ldc_fileserver:$ldc_devshare/efs_qtree},
        qq{Initialized fileqtree $ldc_fileserver:$ldc_distshare/efs_qtree},
        qq{Initialized fileqtree $ldc_fileserver:$ldc_homeshare/efs_qtree},
        q{Initialized all fileqtrees},
    ],
);

$test->efs_test_command(
    command => qq{efs init fileqtree $rdc_name},
    filter  => qr{^ Initial}msx,
    expect  => [
        qq{Initialized fileqtree $rdc_fileserver:$rdc_devshare/efs_qtree},
        qq{Initialized fileqtree $rdc_fileserver:$rdc_distshare/efs_qtree},
        qq{Initialized fileqtree $rdc_fileserver:$rdc_homeshare/efs_qtree},
        q{Initialized all fileqtrees},
    ],
);

$test->efs_test_command(
    command => qq{efs init fileqtree $lpc_name},
    filter  => qr{^ Initial}msx,
    expect  => [
        qq{Initialized fileqtree $lpc_fileserver:$lpc_distshare/efs_qtree},
        q{Initialized all fileqtrees},
    ],
);

# Create and dist the platform and instance releases. The platform
# release will be disted to prod, and instance to dev only.  This will
# allow us to test the differences in behavior of the execlinks
# commands, when dealing with locked or unlocked releases.  We can
# then dist the instance release to prod, and verify that the dist
# command handles release with backwards alias links

dist_release( q{platform}, [ $pl32_1->{name}, $pl64_1->{name} ], q{prod} );
dist_release( q{instance}, [ $in32_1->{name} ], q{dev} );

# Test the maintenance only commands

my @maintenance = (
    qq{efs rename platform $pl32_1->{name} $pl32_1->{newname}},
    qq{efs rename platforminstance $in32_1->{name} $in32_1->{newname}},
    q{efs execlinks any},
    qq{efs execlinks cell $ldc_name},
    qq{efs execlinks fileserver $ldc_name $ldc_fileserver},
    qq{efs execlinks fileshare $ldc_name $ldc_fileserver $ldc_devshare},
    qq{efs execlinks fileqtree $ldc_name $ldc_fileserver $ldc_devshare efs_qtree},
);

foreach my $command ( @maintenance ) {
    $test->efs_test_command(
        name    => qq{Maintenance only: $command},
        command => $command,
        not     => 1,
        expect  => q{This operation can only be run when EFS is in maintenance mode},
    );
}

$test->efs_test_command(
    name    => q{Put EFS into maintenance mode},
    command => q{efs define parameter maintenance_mode true},
);

# test the rename commands

$test->efs_test_command(
    name    => q{Old and new platform names must not match},
    command => qq{efs rename platform $pl32_1->{name} $pl32_1->{name}},
    not     => 1,
    expect  => q{Invalid arguments: new name must be different},
);

$test->efs_test_command(
    name    => q{New name must not be an existing platform},
    command => qq{efs rename platform $pl32_1->{name} $pl64_1->{name}},
    not     => 1,
    expect  => qq{Invalid argument: $pl64_1->{name} is already a platform},
);

$test->efs_test_command(
    name    => q{New name must not be in use},
    command => qq{efs rename platform $pl32_1->{name} $in64_1->{name}},
    not     => 1,
    expect  => qq{Invalid argument: $in64_1->{name} is already a platforminstance},
);

$test->efs_test_command(
    name    => q{Rename the platform},
    command => qq{efs rename platform $pl32_1->{name} $pl32_1->{newname}},
    expect  => [
        qq{Renamed platform $pl32_1->{name} to $pl32_1->{newname}},
    ],
);

$pl32_1->{oldname} = $pl32_1->{name};
$pl32_1->{name} = $pl32_1->{newname};

$test->efs_test_command(
    name    => q{Old and new instance names must not match},
    command => qq{efs rename platforminstance $in32_1->{name} $in32_1->{name}},
    not     => 1,
    expect  => q{Invalid arguments: new name must be different},
);

$test->efs_test_command(
    name    => q{New name must not be an existing platform instance},
    command => qq{efs rename platforminstance $in32_1->{name} $in64_1->{name}},
    not     => 1,
    expect  => qq{Invalid argument: $in64_1->{name} is already a platform instance},
);

$test->efs_test_command(
    name    => q{New name must not be in use},
    command => qq{efs rename platforminstance $in32_1->{name} $pl64_1->{name}},
    not     => 1,
    expect  => qq{Invalid argument: $pl64_1->{name} is already a platform},
);

$test->efs_test_command(
    name    => q{Rename the platform instance},
    command => qq{efs rename platforminstance $in32_1->{name} $in32_1->{newname}},
    expect  => [
        qq{Renamed platform instance $in32_1->{name} to $in32_1->{newname}},
    ],
);

$in32_1->{oldname} = $in32_1->{name};
$in32_1->{name} = $in32_1->{newname};

# Test the execlinks commands

$test->efs_test_command(
    command => qq{efs execlinks fileshare $ldc_name $ldc_fileserver $ldc_devshare},
    expect  => [
        init_fileqtree_output(
            $ldc_name, $ldc_fileserver, $ldc_devshare,
        ),
        execlinks_fileqtree_output(
            $ldc_name, $ldc_fileserver, $ldc_devshare,
            q{dev/testexec/testexec/instance/install},
            q{dev/testexec/testexec/platform/install},
        ),
    ],
);

$test->efs_test_command(
    command => qq{efs execlinks fileserver $ldc_name $ldc_fileserver},
    expect  => [
        qq{Initialized fileqtree $ldc_fileserver:$ldc_devshare/efs_qtree},
        qq{Updated execlinks for $ldc_fileserver:$ldc_devshare/efs_qtree},
        init_fileqtree_output(
            $ldc_name, $ldc_fileserver, $ldc_distshare,
        ),
        execlinks_fileqtree_output(
            $ldc_name, $ldc_fileserver, $ldc_distshare,
            q{dist/testexec/testexec/instance},
            q{dist/testexec/testexec/platform},
        ),
    ],
);

$test->efs_test_command(
    command => qq{efs execlinks cell $rdc_name},
    expect  => [
        init_fileqtree_output(
            $rdc_name, $rdc_fileserver, $rdc_devshare,
        ),
        qq{Updated execlinks for $rdc_fileserver:$rdc_devshare/efs_qtree},
        init_fileqtree_output(
            $rdc_name, $rdc_fileserver, $rdc_distshare,
        ),
        execlinks_fileqtree_output(
            $rdc_name, $rdc_fileserver, $rdc_distshare,
            q{dist/testexec/testexec/instance},
            q{dist/testexec/testexec/platform},
        ),
    ],
);

$test->efs_test_command(
    command => q{efs execlinks any},
    expect  => [
        qq{Initialized fileqtree $ldc_fileserver:$ldc_devshare/efs_qtree},
        qq{Updated execlinks for $ldc_fileserver:$ldc_devshare/efs_qtree},
        qq{Initialized fileqtree $ldc_fileserver:$ldc_distshare/efs_qtree},
        qq{Updated execlinks for $ldc_fileserver:$ldc_distshare/efs_qtree},
        q{Updated execlinks on all fileqtrees},
        qq{Initialized fileqtree $rdc_fileserver:$rdc_devshare/efs_qtree},
        qq{Updated execlinks for $rdc_fileserver:$rdc_devshare/efs_qtree},
        qq{Initialized fileqtree $rdc_fileserver:$rdc_distshare/efs_qtree},
        qq{Updated execlinks for $rdc_fileserver:$rdc_distshare/efs_qtree},
        q{Updated execlinks on all fileqtrees},
        init_fileqtree_output(
            $lpc_name, $lpc_fileserver, $lpc_distshare,
        ),
        execlinks_fileqtree_output(
            $lpc_name, $lpc_fileserver, $lpc_distshare,
            q{dist/testexec/testexec/instance},
            q{dist/testexec/testexec/platform},
        ),
        q{Updated execlinks in all cells},
    ],
);

# Now test the exec control functionality

$test->efs_test_command(
    command => qq{efs create install testexec testexec none $pl32_1->{name}},
    expect  => [ create_install_output( q{none}, $pl32_1->{name} ) ],
);

$test->efs_test_command(
    command => qq{efs create install testexec testexec backwards $pl32_1->{name}},
    expect  => [ create_install_output( q{backwards}, $pl32_1->{name} ) ],
);

$test->efs_test_command(
    command => qq{efs create install testexec testexec emulates $pl32_1->{name}},
    expect  => [ create_install_output( q{emulates}, $pl32_1->{name} ) ],
);

$test->efs_test_command(
    command => qq{efs create install testexec testexec both $pl32_1->{name}},
    expect  => [ create_install_output( q{both}, $pl32_1->{name} ) ],
);

# NOTE: We need to test how the links are created when there is both a
# 32 and 64 bit install, in particular, test how the
# backwards/emulates links are created.  We're going to re-use the
# "both" release, and have that mean two things: both attributes, and
# both installs.

$test->efs_test_command(
    command => qq{efs create install testexec testexec both $pl64_1->{name}},
);

# dist testexec/testexec/none

dist_release( q{none},      [ $pl32_1->{name} ], q{dev} );
dist_release( q{backwards}, [ $pl32_1->{name} ], q{dev} );
dist_release( q{emulates},  [ $pl32_1->{name} ], q{prod} );
dist_release( q{both},      [ $pl32_1->{name},  $pl64_1->{name} ], q{prod} );

# Invert the attribute settings for ALL of these releases

my @setattr = (

    q{efs setattr release testexec testexec none disable_exec_emulates true},
    q{efs setattr release testexec testexec none disable_exec_backwards true},

    q{efs delattr release testexec testexec backwards disable_exec_backwards},
    q{efs setattr release testexec testexec backwards } .
        q{disable_exec_emulates true},

    q{efs delattr release testexec testexec emulates disable_exec_emulates},
    q{efs setattr release testexec testexec emulates } .
        q{disable_exec_backwards true},

    q{efs delattr release testexec testexec both disable_exec_emulates},
    q{efs delattr release testexec testexec both disable_exec_backwards},

);

foreach my $setattr ( @setattr ) {
    $test->efs_test_command( command => $setattr );
}

# test execlinks fileqtree fixing the .exec tree

$test->efs_test_command(
    command => qq{efs execlinks fileqtree $ldc_name $ldc_fileserver $ldc_devshare efs_qtree},
    expect => [
        qq{Initialized fileqtree $ldc_fileserver:$ldc_devshare/efs_qtree},
        execlinks_output_backwards( $ldc_name, q{dev} ),
        execlinks_output_both( $ldc_name, q{dev} ),
        execlinks_output_emulates( $ldc_name, q{dev} ),
        execlinks_output_none( $ldc_name, q{dev} ),
        qq{Updated execlinks for $ldc_fileserver:$ldc_devshare/efs_qtree},
    ],
);

# test execlinks mpr commands fixing the .exec tree

$test->efs_test_command(
    command => qq{efs execlinks release testexec testexec emulates -cells $lpc_name},
    expect => [
        q{Updated execlinks for 'testexec/testexec/emulates' in /efs/dev},
        execlinks_output_emulates( $lpc_name, q{dist} ),
        q{Updated execlinks for 'testexec/testexec/emulates' in /efs/dist},
    ],
);

$test->efs_test_command(
    command => qq{efs execlinks release testexec testexec both -cells $lpc_name},
    expect => [
        q{Updated execlinks for 'testexec/testexec/both' in /efs/dev},
        execlinks_output_both( $lpc_name, q{dist} ),
        q{Updated execlinks for 'testexec/testexec/both' in /efs/dist},
    ],
);

$test->efs_test_command(
    command => qq{efs execlinks project testexec testexec -cells $rdc_name},
    expect => [
        q{Updating execlinks for project 'testexec/testexec'},
        q{Updated execlinks for 'testexec/testexec/backwards' in /efs/dev},
        execlinks_output_backwards( $rdc_name, q{dist} ),
        q{Updated execlinks for 'testexec/testexec/backwards' in /efs/dist},
        q{Updated execlinks for 'testexec/testexec/both' in /efs/dev},
        execlinks_output_both( $rdc_name, q{dist} ),
        q{Updated execlinks for 'testexec/testexec/both' in /efs/dist},
        q{Updated execlinks for 'testexec/testexec/emulates' in /efs/dev},
        execlinks_output_emulates( $rdc_name, q{dist} ),
        q{Updated execlinks for 'testexec/testexec/emulates' in /efs/dist},
        q{Updated execlinks for 'testexec/testexec/instance' in /efs/dev},
        q{Updated execlinks for 'testexec/testexec/instance' in /efs/dist},
        q{Updated execlinks for 'testexec/testexec/none' in /efs/dev},
        execlinks_output_none( $rdc_name, q{dist} ),
        q{Updated execlinks for 'testexec/testexec/none' in /efs/dist},
        q{Updated execlinks for 'testexec/testexec/platform' in /efs/dev},
        q{Updated execlinks for 'testexec/testexec/platform' in /efs/dist},
        q{Successfully updated all execlinks for project 'testexec/testexec'},
    ],
);

$test->efs_test_command(
    command => qq{efs execlinks metaproj testexec -cells $ldc_name},
    expect => [
        q{Updating execlinks in entire metaproj 'testexec'},
        q{Updating execlinks for project 'testexec/testexec'},
        q{Updated execlinks for 'testexec/testexec/backwards' in /efs/dev},
        execlinks_output_backwards( $ldc_name, q{dist} ),
        q{Updated execlinks for 'testexec/testexec/backwards' in /efs/dist},
        q{Updated execlinks for 'testexec/testexec/both' in /efs/dev},
        execlinks_output_both( $ldc_name, q{dist} ),
        q{Updated execlinks for 'testexec/testexec/both' in /efs/dist},
        q{Updated execlinks for 'testexec/testexec/emulates' in /efs/dev},
        execlinks_output_emulates( $ldc_name, q{dist} ),
        q{Updated execlinks for 'testexec/testexec/emulates' in /efs/dist},
        q{Updated execlinks for 'testexec/testexec/instance' in /efs/dev},
        q{Updated execlinks for 'testexec/testexec/instance' in /efs/dist},
        q{Updated execlinks for 'testexec/testexec/none' in /efs/dev},
        execlinks_output_none( $ldc_name, q{dist} ),
        q{Updated execlinks for 'testexec/testexec/none' in /efs/dist},
        q{Updated execlinks for 'testexec/testexec/platform' in /efs/dev},
        q{Updated execlinks for 'testexec/testexec/platform' in /efs/dist},
        q{Successfully updated all execlinks for project 'testexec/testexec'},
        q{Successfully updated all execlinks for metaproj 'testexec'},
    ],
);

$test->efs_test_command(
    name    => q{Take EFS out of maintenance mode},
    command => q{efs undef parameter maintenance_mode},
);

# Test create install behavior when alias install exists

$test->efs_test_command(
    command => q{efs unlock release testexec testexec instance},
);

$test->efs_test_link(
    link   => qq{/efs/dev/testexec/testexec/instance/install/.exec/$in32_1->{name}},
    target => $in32_1->{oldname},
);

$test->efs_test_command(
    command => qq{efs create install testexec testexec instance $in32_1->{name}},
    not     => 1,
    expect  => [
        q{This release has one or more alias } .
            qq{installs for platform(instance) $in32_1->{name}:},
        qq{testexec/testexec/instance/$in32_1->{oldname}},
        q{Use --force if you want both types},
    ],
);

$test->efs_test_command(
    command => qq{efs create install testexec testexec instance $in32_1->{name} --force},
);

$test->efs_test_directory(
    nolink    => 1,
    directory => qq{/efs/dev/testexec/testexec/instance/install/.exec/$in32_1->{name}},
    user      => q{efsops},
    group     => q{efsops},
);

$test->efs_test_command(
    command => qq{efs create install testexec testexec instance $pl32_1->{oldname}},
    filter  => qr{Normalizing}msx,
    expect  => qq{Normalizing platform alias $pl32_1->{oldname} to $pl32_1->{name}},
);

# Test adding a 2nd, newer platform to an existing release.

$test->efs_test_command(
    command => q{efs unlock release testexec testexec platform},
);

$test->efs_test_link(
    link   => qq{/efs/dev/testexec/testexec/platform/install/.exec/$pl32_2->{name}},
    target => $pl32_1->{oldname},
);

$test->efs_test_link(
    link   => qq{/efs/dist/testexec/testexec/platform/.exec/$pl32_2->{name}},
    target => $pl32_1->{oldname},
);

$test->efs_test_command(
    command => qq{efs create install testexec testexec platform $pl32_2->{name}},
);

$test->efs_test_link(
    not    => 1,
    link   => qq{/efs/dev/testexec/testexec/platform/install/.exec/$pl32_2->{name}},
    target => $pl32_1->{name},
);

$test->efs_test_directory(
    nolink    => 1,
    directory => qq{/efs/dev/testexec/testexec/platform/install/.exec/$pl32_2->{name}},
    user      => q{efsops},
    group     => q{efsops},
);

dist_release( 'platform', [ $pl32_2->{name} ], q{dev} );

$test->efs_test_link(
    not    => 1,
    link   => qq{/efs/dist/testexec/testexec/platform/.exec/$pl32_2->{name}},
    target => $pl32_1->{name},
);

$test->efs_test_directory(
    nolink    => 1,
    directory => qq{/efs/dist/testexec/testexec/platform/.exec/$pl32_2->{name}},
    user      => q{efsops},
    group     => q{efsops},
);

# teardown the exec control releases first

my @teardown = ();

foreach my $release ( qw( none backwards emulates both ) ) {
    my $stage = $release ~~ [qw(emulates both)] ? q{deprecated} : q{rejected};
    push @teardown, (
        qq{efs setstage release testexec testexec $release $stage},
        qq{efs destroy release testexec testexec $release},
        qq{efs purge release testexec testexec $release --now},
    );
}

foreach my $teardown ( @teardown ) {
    $test->efs_test_command(
        name    => qq{(teardown, phase one) $teardown},
        command => $teardown,
    );
}

# Test the the undef scenarios

$test->efs_test_command(
    name    => q{Unable to undef a platform instance while it has aliases},
    command => qq{efs undef platforminstance $in64_2->{name}},
    not     => 1,
    expect  => [
        qq{Unable to undef platform instance '$in64_2->{name}',},
        q{as it has the following aliases:},
        qq{$ia64_2->{name}},
    ],
);

$test->efs_test_command(
    name    => q{Unable to undef a platform instance while it has forwards},
    command => qq{efs undef platforminstance $in64_1->{name}},
    not     => 1,
    expect  => [
        qq{Unable to undef platform instance '$in64_1->{name}',},
        q{as it has the following forwards:},
        qq{$in64_2->{name}},
    ],
);

$test->efs_test_command(
    name    => q{Unable to undef a platform instance while it has emulators},
    command => qq{efs undef platforminstance $in32_2->{name}},
    not     => 1,
    expect  => [
        qq{Unable to undef platform instance '$in32_2->{name}',},
        q{as it has the following emulators:},
        qq{$in64_2->{name}},
    ],
);

$test->efs_test_command(
    command => qq{efs undef platforminstancealias $ia64_2->{name}},
    expect  => qq{Undefined platform instance alias $ia64_2->{name}},
);

foreach my $name ( $in64_2->{name}, $in64_1->{name}, $in32_2->{name} ) {
    $test->efs_test_command(
        command => qq{efs undef platforminstance $name},
        expect  => qq{Undefined platform instance $name},
    );
}

$test->efs_test_command(
    name    => q{Unable to undef platforminstance while it has installs},
    command => qq{efs undef platforminstance $in32_1->{name}},
    not     => 1,
    expect  => [
        qq{Unable to undef platform instance '$in32_1->{name}',},
        q{as the following installs use this name:},
        qq{testexec/testexec/instance/$in32_1->{name}},
    ],
);

$test->efs_test_command(
    name    => q{Unable to undef platforminstancealias while it has installs},
    command => qq{efs undef platforminstancealias $in32_1->{oldname}},
    not     => 1,
    expect  => [
        qq{Unable to undef platform instance alias '$in32_1->{oldname}',},
        q{as the following installs use this name:},
        qq{testexec/testexec/instance/$in32_1->{oldname}},
    ],
);

destroy_release( q{instance}, q{rejected} );

$test->efs_test_command(
    command => qq{efs undef platforminstancealias $in32_1->{oldname}},
);

$test->efs_test_command(
    command => qq{efs undef platforminstance $in32_1->{name}},
);

$test->efs_test_command(
    name    => q{Unable to undef platform while it has installs},
    command => qq{efs undef platform $pl32_2->{name}},
    not     => 1,
    expect  => [
        qq{Unable to undef platform '$pl32_2->{name}',},
        q{as the following installs use this name:},
        qq{testexec/testexec/platform/$pl32_2->{name}},
    ],
);

$test->efs_test_command(
    name    => q{Unable to undef platformalias while it has installs},
    command => qq{efs undef platformalias $pl32_1->{oldname}},
    not     => 1,
    expect  => [
        qq{Unable to undef platform alias '$pl32_1->{oldname}',},
        q{as the following installs use this name:},
        qq{testexec/testexec/platform/$pl32_1->{oldname}},
    ],
);

destroy_release( q{platform}, q{deprecated} );

$test->efs_test_command(
    name    => q{Unable to undef a platform while it has aliases},
    command => qq{efs undef platform $pl64_2->{name}},
    not     => 1,
    expect  => [
        qq{Unable to undef platform '$pl64_2->{name}',},
        q{as it has the following aliases:},
        qq{$pa64_2->{name}},
    ],
);

$test->efs_test_command(
    name    => q{Unable to undef a platform while it has forwards},
    command => qq{efs undef platform $pl64_1->{name}},
    not     => 1,
    expect  => [
        qq{Unable to undef platform '$pl64_1->{name}',},
        q{as it has the following forwards:},
        qq{$pl64_2->{name}},
    ],
);

$test->efs_test_command(
    name    => q{Unable to undef a platform while it has emulators},
    command => qq{efs undef platform $pl32_2->{name}},
    not     => 1,
    expect  => [
        qq{Unable to undef platform '$pl32_2->{name}',},
        q{as it has the following emulators:},
        qq{$pl64_2->{name}},
    ],
);

$test->efs_test_command(
    command => qq{efs undef platformalias $pa64_2->{name}},
    expect  => qq{Undefined platform alias $pa64_2->{name}},
);

foreach my $name ( $pl64_2->{name}, $pl64_1->{name}, $pl32_2->{name} ) {
    $test->efs_test_command(
        command => qq{efs undef platform $name},
        expect  => qq{Undefined platform $name},
    );
}

$test->efs_test_command(
    command => qq{efs undef platformalias $pl32_1->{oldname}},
);

$test->efs_test_command(
    command => qq{efs undef platform $pl32_1->{name}},
);

# Add history tests for all 4 object types

my @history = (

    {
        name   => $pl32_1->{oldname},
        type   => q{platform},
        expect => [
            qr{ define \s+ platform \s+ $pl32_1->{oldname} }msx,
            qr{ define \s+ platforminstance \s+ $pl32_1->{oldname} \s+ $pl32_1->{oldname} }msx,
            qr{ define \s+ platforminstance \s+ $in32_1->{oldname} \s+ $pl32_1->{oldname} \s+ nosuchinstance }msx,
            qr{ define \s+ platforminstance \s+ $in32_1->{oldname} \s+ $pl32_1->{oldname} \s+ --emulates \s+ nosuchinstance }msx,
            qr{ define \s+ platforminstance \s+ $in32_1->{oldname} \s+ $pl32_1->{oldname} \s+ $in32_1->{oldname} }msx,
            qr{ define \s+ platforminstance \s+ $in32_1->{oldname} \s+ $pl32_1->{oldname} \s+ --emulates \s+ $in32_1->{oldname} }msx,
            qr{ define \s+ platforminstance \s+ $in32_1->{oldname} \s+ $pl32_1->{oldname} }msx,
            qr{ rename \s+ platform \s+ $pl32_1->{oldname} \s+ $pl32_1->{name} }msx,
            qr{ rename \s+ platform \s+ $pl32_1->{oldname} \s+ $pl32_1->{oldname} }msx,
            qr{ rename \s+ platform \s+ $pl32_1->{oldname} \s+ $pl64_1->{name} }msx,
            qr{ rename \s+ platform \s+ $pl32_1->{oldname} \s+ $in64_1->{name} }msx,
            qr{ rename \s+ platform \s+ $pl32_1->{oldname} \s+ $pl32_1->{name} }msx,
        ],
    },

    {
        name   => $pl32_1->{oldname},
        type   => q{platformalias},
        expect => [
            qr{ undef \s+ platformalias \s+ $pl32_1->{oldname} }msx,
            qr{ undef \s+ platformalias \s+ $pl32_1->{oldname} }msx,
        ],
    },

    {
        name   => $pl32_1->{name},
        type   => q{platform},
        expect => [
            qr{ undef \s+ platform \s+ $pl32_1->{name} }msx,
        ],
    },

    {
        name   => $pl32_2->{name},
        type   => q{platform},
        expect => [
            qr{ define \s+ platform \s+ $pl32_2->{name} }msx,
            qr{ define \s+ platform \s+ $pl32_2->{name} \s+ $pl32_1->{oldname} }msx,
            qr{ define \s+ platforminstance \s+ $in32_2->{name} \s+ $pl32_2->{name} }msx,
            qr{ define \s+ platforminstance \s+ $in32_2->{name} \s+ $pl32_2->{name} \s+ $in32_1->{oldname} }msx,
            qr{ undef \s+ platform \s+ $pl32_2->{name} }msx,
            qr{ undef \s+ platform \s+ $pl32_2->{name} }msx,
            qr{ undef \s+ platform \s+ $pl32_2->{name} }msx,
        ],
    },

    {
        name   => $pl64_1->{name},
        type   => q{platform},
        expect => [
            qr{ define \s+ platform \s+ $pl64_1->{name} \s+ nosuchplatform }msx,
            qr{ define \s+ platform \s+ $pl64_1->{name} \s+ --emulates \s+ nosuchplatform }msx,
            qr{ define \s+ platform \s+ $pl64_1->{name} \s+ $pl64_1->{name} }msx,
            qr{ define \s+ platform \s+ $pl64_1->{name} \s+ --emulates \s+ $pl64_1->{name} }msx,
            qr{ define \s+ platform \s+ $pl64_1->{name} }msx,
            qr{ define \s+ platform \s+ $pl64_1->{name} \s+ --emulates \s+ $pl32_1->{oldname} }msx,
            qr{ define \s+ platforminstance \s+ $in64_1->{name} \s+ $pl64_1->{name} }msx,
            qr{ define \s+ platforminstance \s+ $in64_1->{name} \s+ $pl64_1->{name} \s+ --emulates \s+ $in32_1->{oldname} }msx,
            qr{ undef \s+ platform \s+ $pl64_1->{name} }msx,
            qr{ undef \s+ platform \s+ $pl64_1->{name} }msx,
        ],
    },

    {
        name   => $pl64_2->{name},
        type   => q{platform},
        expect => [
            qr{ define \s+ platform \s+ $pl64_2->{name} \s+ $pl64_1->{name} \s+ $pl32_2->{name} }msx,
            qr{ define \s+ platforminstance \s+ $in64_2->{name} \s+ $pl64_2->{name} \s+ $in64_1->{name} \s+ $in32_2->{name} }msx,
            qr{ define \s+ platformalias \s+ $pl64_2->{name} \s+ $pl64_2->{name} }msx,
            qr{ define \s+ platformalias \s+ $pa64_2->{name} \s+ $pl64_2->{name} }msx,
            qr{ undef \s+ platform \s+ $pl64_2->{name} }msx,
            qr{ undef \s+ platform \s+ $pl64_2->{name} }msx,
        ],
    },

    {
        name   => $pa64_2->{name},
        type   => q{platformalias},
        expect => [
            qr{ efsops \s+ define \s+ platformalias \s+ $pa64_2->{name} \s+ nosuchplatform }msx,
            qr{ define \s+ platformalias \s+ $pa64_2->{name} \s+ $pl64_2->{name} }msx,
            qr{ undef \s+ platformalias \s+ $pa64_2->{name} }msx,
        ],
    },

    {
        name   => $in32_1->{oldname},
        type   => q{platforminstance},
        expect => [
            qr{ define \s+ platforminstance \s+ $in32_1->{oldname} \s+ nosuchplatform }msx,
            qr{ define \s+ platforminstance \s+ $in32_1->{oldname} \s+ $pl32_1->{oldname} \s+ nosuchinstance }msx,
            qr{ define \s+ platforminstance \s+ $in32_1->{oldname} \s+ $pl32_1->{oldname} \s+ --emulates \s+ nosuchinstance }msx,
            qr{ define \s+ platforminstance \s+ $in32_1->{oldname} \s+ $pl32_1->{oldname} \s+ $in32_1->{oldname} }msx,
            qr{ define \s+ platforminstance \s+ $in32_1->{oldname} \s+ $pl32_1->{oldname} \s+ --emulates \s+ $in32_1->{oldname} }msx,
            qr{ define \s+ platforminstance \s+ $in32_1->{oldname} \s+ $pl32_1->{oldname} }msx,
            qr{ rename \s+ platforminstance \s+ $in32_1->{oldname} \s+ $in32_1->{name} }msx,
            qr{ rename \s+ platforminstance \s+ $in32_1->{oldname} \s+ $in32_1->{oldname} }msx,
            qr{ rename \s+ platforminstance \s+ $in32_1->{oldname} \s+ $in64_1->{name} }msx,
            qr{ rename \s+ platforminstance \s+ $in32_1->{oldname} \s+ $pl64_1->{name} }msx,
            qr{ rename \s+ platforminstance \s+ $in32_1->{oldname} \s+ $in32_1->{name} }msx,
        ],
    },

    {
        name   => $in32_1->{oldname},
        type   => q{platforminstancealias},
        expect => [
            qr{ define \s+ platforminstancealias \s+ $in32_1->{oldname} \s+ $in64_2->{name} }msx,
            qr{ undef \s+ platforminstancealias \s+ $in32_1->{oldname} }msx,
            qr{ undef \s+ platforminstancealias \s+ $in32_1->{oldname} }msx,
        ],
    },

    {
        name   => $in32_1->{name},
        type   => q{platforminstance},
        expect => [
            qr{ undef \s+ platforminstance \s+ $in32_1->{name} }msx,
            qr{ undef \s+ platforminstance \s+ $in32_1->{name} }msx,
        ],
    },

    {
        name   => $in32_2->{name},
        type   => q{platforminstance},
        expect => [
            qr{ define \s+ platforminstance \s+ $in32_2->{name} \s+ $pl32_2->{name} }msx,
            qr{ define \s+ platforminstance \s+ $in32_2->{name} \s+ $pl32_2->{name} \s+ $in32_1->{oldname} }msx,
            qr{ undef \s+ platforminstance \s+ $in32_2->{name} }msx,
            qr{ undef \s+ platforminstance \s+ $in32_2->{name} }msx,
        ],
    },

    {
        name   => $in64_1->{name},
        type   => q{platforminstance},
        expect => [
            qr{ define \s+ platforminstance \s+ $in64_1->{name} \s+ $pl64_1->{name} }msx,
            qr{ define \s+ platforminstance \s+ $in64_1->{name} \s+ $pl64_1->{name} \s+ --emulates \s+ $in32_1->{oldname} }msx,
            qr{ undef \s+ platforminstance \s+ $in64_1->{name} }msx,
            qr{ undef \s+ platforminstance \s+ $in64_1->{name} }msx,
        ],
    },

    {
        name   => $in64_2->{name},
        type   => q{platforminstance},
        expect => [
            qr{ define \s+ platforminstance \s+ $in64_2->{name} \s+ $pl64_2->{name} \s+ $in64_1->{name} \s+ $in32_2->{name} }msx,
            qr{ define \s+ platforminstancealias \s+ $in32_1->{oldname} \s+ $in64_2->{name} }msx,
            qr{ define \s+ platforminstancealias \s+ $ia64_2->{name} \s+ $in64_2->{name} }msx,
            qr{ undef \s+ platforminstance \s+ $in64_2->{name} }msx,
            qr{ undef \s+ platforminstance \s+ $in64_2->{name} }msx,
        ],
    },

    {
        name   => $ia64_2->{name},
        type   => q{platforminstancealias},
        expect => [
            qr{ define \s+ platforminstancealias \s+ $ia64_2->{name} \s+ nosuchinstance }msx,
            qr{ define \s+ platforminstancealias \s+ $ia64_2->{name} \s+ $in64_2->{name} }msx,
            qr{ undef \s+ platforminstancealias \s+ $ia64_2->{name} }msx,
        ],
    },

);

foreach my $history ( @history ) {

    my $count = scalar @{ $history->{expect} };

    $test->efs_test_command(
        command => qq{efs history $history->{type} $history->{name} --count $count},
        filter  => qr{ \s+ efsops \s+ }msx,
        expect  => $history->{expect},
    );

}

$test->efs_test_command(
    name    => q{Put EFS into maintenance mode},
    command => q{efs define parameter maintenance_mode true},
);

$test->efs_test_command(
    name    => qq{Reinit all fileqtrees to remove undefined instances, $ldc_name},
    command => qq{efs init fileqtree $ldc_name},
    expect  => [
        teardown_fileqtree_output( $ldc_name, $ldc_fileserver, $ldc_devshare ),
        teardown_fileqtree_output( $ldc_name, $ldc_fileserver, $ldc_distshare ),
        teardown_fileqtree_output( $ldc_name, $ldc_fileserver, $ldc_homeshare ),
        q{Initialized all fileqtrees},
    ],
);

$test->efs_test_command(
    name    => qq{Reinit all fileqtrees to remove undefined instances, $rdc_name},
    command => qq{efs init fileqtree $rdc_name},
    expect  => [
        teardown_fileqtree_output( $rdc_name, $rdc_fileserver, $rdc_devshare ),
        teardown_fileqtree_output( $rdc_name, $rdc_fileserver, $rdc_distshare ),
        teardown_fileqtree_output( $rdc_name, $rdc_fileserver, $rdc_homeshare ),
        q{Initialized all fileqtrees},
    ],
);

$test->efs_test_command(
    name    => qq{Reinit all fileqtrees to remove undefined instances, $lpc_name},
    command => qq{efs init fileqtree $lpc_name},
    expect  => [
        teardown_fileqtree_output( $lpc_name, $lpc_fileserver, $lpc_distshare ),
        q{Initialized all fileqtrees},
    ],
);

$test->efs_test_command(
    name    => q{Take EFS out of maintenance mode},
    command => q{efs undef parameter maintenance_mode},
);

@teardown = (
    q{efs destroy project testexec testexec},
    q{efs destroy metaproj testexec},
);

foreach my $teardown ( @teardown ) {
    $test->efs_test_command(
        name    => qq{(teardown, phase two) $teardown},
        command => $teardown,
    );
}

done_testing();

exit 0;

sub destroy_release {

    my ($release,$stage) = @_;

    my @commands = (
        qq{efs setstage release testexec testexec $release $stage},
        qq{efs destroy release testexec testexec $release},
        qq{efs purge release testexec testexec $release -now},
    );

    foreach my $command ( @commands ) {
        $test->efs_test_command( command => $command );
    }

    return;

}

sub dist_release {

    my ($release,$installs,$celltype) = @_;

    my @commands = ();

    foreach my $install ( @{ $installs } ) {
        push @commands, (
            qq{efs create install testexec testexec $release $install},
            qq{touch /efs/dev/testexec/testexec/$release/install/.exec/$install/foo},
        );
    }

    if ( $celltype eq q{prod} ) {
        push @commands,
            qq{efs checkpoint release testexec testexec $release};
    } else {
        push @commands,
            qq{efs checkpoint release testexec testexec $release --nolock};
    }

    push @commands, qq{efs dist release testexec testexec $release --global};

    if ( $celltype eq q{prod} ) {
        push @commands, (
            qq{efs setstage release testexec testexec $release prod},
            qq{efs dist release testexec testexec $release --celltype prod --global},
        );
    }

    foreach my $command ( @commands ) {
        $test->efs_test_command( command => $command );
    }

    return;

}

sub init_fileqtree_output {

    my ($cell,$fileserver,$fileshare) = @_;
    return (
        qq{Renamed $cell:/.efs/$in32_1->{oldname} } .
            qq{to $cell:/.efs/$in32_1->{name}},
        qq{Created symlink $cell:/.efs/$in32_1->{oldname} -> $in32_1->{name}},
        qq{Renamed $cell:/.efs/$pl32_1->{oldname} } .
            qq{to $cell:/.efs/$pl32_1->{name}},
        qq{Created symlink $cell:/.efs/$pl32_1->{oldname} -> $pl32_1->{name}},
        qq{Initialized fileqtree $fileserver:$fileshare/efs_qtree},
    );

}

sub teardown_fileqtree_output {

    my ($cell,$fileserver,$fileshare) = @_;

    if ( $fileshare =~ m{efs_home}msx ) {
        return (
            qq{Initialized fileqtree $fileserver:$fileshare/efs_qtree},
        );
    } else {
        return (
            qq{Removed directory $cell:/.efs/$in32_1->{name}},
            qq{Removed symlink $cell:/.efs/$in32_1->{oldname}},
            qq{Removed directory $cell:/.efs/$in32_2->{name}},
            qq{Removed symlink $cell:/.efs/$ia64_2->{name}},
            qq{Removed directory $cell:/.efs/$in64_1->{name}},
            qq{Removed directory $cell:/.efs/$in64_2->{name}},
            qq{Removed directory $cell:/.efs/$pl32_1->{name}},
            qq{Removed symlink $cell:/.efs/$pl32_1->{oldname}},
            qq{Removed directory $cell:/.efs/$pl32_2->{name}},
            qq{Removed symlink $cell:/.efs/$pa64_2->{name}},
            qq{Removed directory $cell:/.efs/$pl64_1->{name}},
            qq{Removed directory $cell:/.efs/$pl64_2->{name}},
            qq{Initialized fileqtree $fileserver:$fileshare/efs_qtree},
        );
    }

}

sub execlinks_output_backwards {

    my ($cell,$type) = @_;

    my $path = "$type/testexec/testexec/backwards";
    $path    .= q{/install} if $type eq q{dev};

    my @results = (
        qq{Created symlink $cell:/efs/$path/.exec/$in32_2->{name} -> $pl32_1->{name}},
        qq{Created symlink $cell:/efs/$path/.exec/$pl32_2->{name} -> $pl32_1->{name}},
        qq{Removed symlink $cell:/efs/$path/.exec/$in64_1->{name}},
        qq{Removed symlink $cell:/efs/$path/.exec/$pl64_1->{name}},
    );

    if ( $type eq q{dev} ) {
        my $dir = dirname $path;
        push @results, qq{Created directory $cell:/.efs/$in32_2->{name}/$dir};
    }

    push @results, (
        qq{Created directory $cell:/.efs/$in32_2->{name}/$path},
        qq{Created symlink $cell:/.efs/$in32_2->{name}/$path/exec } .
            qq{-> /efs/$path/.exec/$in32_2->{name}},
    );

    if ( $type eq q{dev} ) {
        my $dir = dirname $path;
        push @results, qq{Created directory $cell:/.efs/$pl32_2->{name}/$dir};
    }

    push @results, (
        qq{Created directory $cell:/.efs/$pl32_2->{name}/$path},
        qq{Created symlink $cell:/.efs/$pl32_2->{name}/$path/exec } .
            qq{-> /efs/$path/.exec/$pl32_2->{name}},
    );

    return @results;

}

sub execlinks_output_both {

    my ($cell,$type) = @_;

    my $path = "$type/testexec/testexec/both";
    $path    .= q{/install} if $type eq q{dev};

    my @results = (
        qq{Created symlink $cell:/efs/$path/.exec/$in32_2->{name} -> $pl32_1->{name}},
        qq{Created symlink $cell:/efs/$path/.exec/$ia64_2->{name} -> $pl64_1->{name}},
        qq{Created symlink $cell:/efs/$path/.exec/$in64_2->{name} -> $pl64_1->{name}},
        qq{Created symlink $cell:/efs/$path/.exec/$pl32_2->{name} -> $pl32_1->{name}},
        qq{Created symlink $cell:/efs/$path/.exec/$pa64_2->{name} -> $pl64_1->{name}},
        qq{Created symlink $cell:/efs/$path/.exec/$pl64_2->{name} -> $pl64_1->{name}},
    );

    foreach my $name ( $in32_2->{name}, $in64_2->{name},
                       $pl32_2->{name}, $pl64_2->{name} ) {

        if ( $type eq q{dev} ) {
            my $dir = dirname $path;
            push @results,
                qq{Created directory $cell:/.efs/$name/$dir};
        }

        push @results, (
            qq{Created directory $cell:/.efs/$name/$path},
            qq{Created symlink $cell:/.efs/$name/$path/exec -> /efs/$path/.exec/$name},
        );

    }

    return @results;

}

sub execlinks_output_emulates {

    my ($cell,$type) = @_;

    my $path = "$type/testexec/testexec/emulates";
    $path    .= q{/install} if $type eq q{dev};

    my @results = (
        qq{Created symlink $cell:/efs/$path/.exec/$in64_1->{name} -> $pl32_1->{name}},
        qq{Created symlink $cell:/efs/$path/.exec/$pl64_1->{name} -> $pl32_1->{name}},
    );

    if ( $type eq q{dev} ) {
        my $dir = dirname $path;
        push @results,
            qq{Created directory $cell:/.efs/$in64_1->{name}/$dir};
    }

    push @results, (
        qq{Created directory $cell:/.efs/$in64_1->{name}/$path},
        qq{Created symlink $cell:/.efs/$in64_1->{name}/$path/exec -> /efs/$path/.exec/$in64_1->{name}},
    );

    if ( $type eq q{dev} ) {
        my $dir = dirname $path;
        push @results,
            qq{Created directory $cell:/.efs/$pl64_1->{name}/$dir};
    }

    push @results, (
        qq{Created directory $cell:/.efs/$pl64_1->{name}/$path},
        qq{Created symlink $cell:/.efs/$pl64_1->{name}/$path/exec -> /efs/$path/.exec/$pl64_1->{name}},
    );

    return @results;

}

sub execlinks_output_none {

    my ($cell,$type) = @_;

    my $path = "$type/testexec/testexec/none";
    $path    .= q{/install} if $type eq q{dev};

    my @results = (
        qq{Removed symlink $cell:/efs/$path/.exec/$in32_2->{name}},
        qq{Removed symlink $cell:/efs/$path/.exec/$ia64_2->{name}},
        qq{Removed symlink $cell:/efs/$path/.exec/$in64_1->{name}},
        qq{Removed symlink $cell:/efs/$path/.exec/$in64_2->{name}},
        qq{Removed symlink $cell:/efs/$path/.exec/$pl32_2->{name}},
        qq{Removed symlink $cell:/efs/$path/.exec/$pa64_2->{name}},
        qq{Removed symlink $cell:/efs/$path/.exec/$pl64_1->{name}},
        qq{Removed symlink $cell:/efs/$path/.exec/$pl64_2->{name}},
    );

    return @results;

}

sub execlinks_fileqtree_output {

    my ( $cell, $fileserver, $fileshare,
         $instance_path, $platform_path ) = @_;

    my $instance_dotexec = qq{/efs/$instance_path/.exec};
    my $platform_dotexec = qq{/efs/$platform_path/.exec};

    my @results = ();

    if ( $cell ne $lpc_name ) {
        push @results, (
            qq{Created symlink $cell:$instance_dotexec/$in32_1->{name} } .
                qq{-> $in32_1->{oldname}},
            q{Changed symlink } .
                qq{$cell:/.efs/$in32_1->{name}/$instance_path/exec from } .
                    qq{$instance_dotexec/$in32_1->{oldname} to } .
                        qq{$instance_dotexec/$in32_1->{name}},
        );
    }

    push @results, (
        qq{Created symlink $cell:$platform_dotexec/$in32_1->{name} } .
            qq{-> $pl32_1->{oldname}},
        qq{Created symlink $cell:$platform_dotexec/$pl32_1->{name} } .
            qq{-> $pl32_1->{oldname}},
        qq{Updated execlinks for $fileserver:$fileshare/efs_qtree},
        q{Updated execlinks on all fileqtrees},
    );

    return @results;

}

sub create_install_output {

    my $release = shift;
    my $install = shift;

    my $bitness = $install =~ m{64}ms ? 64 : 32;

    my $dev_release = qq{dev/testexec/testexec/$release};
    my $dev_install = qq{$dev_release/install};

    my @data = (
        [
            qq{Created directory $ldc_name:/efs/$dev_install/.exec},
            qq{Created directory $ldc_name:/efs/$dev_install/.exec/$pl32_1->{name}},
            qq{Changed ownership of $ldc_name:/efs/$dev_install/.exec/$pl32_1->{name} to efsops/efsops},
            qq{Created symlink $ldc_name:/efs/$dev_install/.exec/$in32_1->{name} -> $pl32_1->{name}},
            qq{Created symlink $ldc_name:/efs/$dev_install/.exec/$in32_1->{oldname} -> $pl32_1->{name}},
        ],
        [
            [qw( none emulates )],
            qq{Created symlink $ldc_name:/efs/$dev_install/.exec/$in32_2->{name} -> $pl32_1->{name}},
        ],
        [
            [qw( none )],
            qq{Created symlink $ldc_name:/efs/$dev_install/.exec/$ia64_2->{name} -> $pl32_1->{name}},
        ],
        [
            [qw( none backwards )],
            qq{Created symlink $ldc_name:/efs/$dev_install/.exec/$in64_1->{name} -> $pl32_1->{name}},
        ],
        [
            [qw( none )],
            qq{Created symlink $ldc_name:/efs/$dev_install/.exec/$in64_2->{name} -> $pl32_1->{name}},
        ],
        [
            qq{Created symlink $ldc_name:/efs/$dev_install/.exec/$pl32_1->{oldname} -> $pl32_1->{name}},
        ],
        [
            [qw( none emulates )],
            qq{Created symlink $ldc_name:/efs/$dev_install/.exec/$pl32_2->{name} -> $pl32_1->{name}},
        ],
        [
            [qw( none )],
            qq{Created symlink $ldc_name:/efs/$dev_install/.exec/$pa64_2->{name} -> $pl32_1->{name}},
        ],
        [
            [qw( none backwards )],
            qq{Created symlink $ldc_name:/efs/$dev_install/.exec/$pl64_1->{name} -> $pl32_1->{name}},
        ],
        [
            [qw( none )],
            qq{Created symlink $ldc_name:/efs/$dev_install/.exec/$pl64_2->{name} -> $pl32_1->{name}},
        ],
        [
            qq{Created directory $ldc_name:/.efs/$in32_1->{name}/$dev_release},
            qq{Created directory $ldc_name:/.efs/$in32_1->{name}/$dev_install},
            qq{Created symlink $ldc_name:/.efs/$in32_1->{name}/$dev_install/exec -> /efs/$dev_install/.exec/$in32_1->{name}},
        ],
        [
            [qw( none emulates )],
            qq{Created directory $ldc_name:/.efs/$in32_2->{name}/$dev_release},
            qq{Created directory $ldc_name:/.efs/$in32_2->{name}/$dev_install},
            qq{Created symlink $ldc_name:/.efs/$in32_2->{name}/$dev_install/exec -> /efs/$dev_install/.exec/$in32_2->{name}},
        ],
        [
            [qw( none backwards)],
            qq{Created directory $ldc_name:/.efs/$in64_1->{name}/$dev_release},
            qq{Created directory $ldc_name:/.efs/$in64_1->{name}/$dev_install},
            qq{Created symlink $ldc_name:/.efs/$in64_1->{name}/$dev_install/exec -> /efs/$dev_install/.exec/$in64_1->{name}},
        ],
        [
            [qw( none )],
            qq{Created directory $ldc_name:/.efs/$in64_2->{name}/$dev_release},
            qq{Created directory $ldc_name:/.efs/$in64_2->{name}/$dev_install},
            qq{Created symlink $ldc_name:/.efs/$in64_2->{name}/$dev_install/exec -> /efs/$dev_install/.exec/$in64_2->{name}},
        ],
        [
            qq{Created directory $ldc_name:/.efs/$pl32_1->{name}/$dev_release},
            qq{Created directory $ldc_name:/.efs/$pl32_1->{name}/$dev_install},
            qq{Created symlink $ldc_name:/.efs/$pl32_1->{name}/$dev_install/exec -> /efs/$dev_install/.exec/$pl32_1->{name}},
        ],
        [
            [qw( none emulates )],
            qq{Created directory $ldc_name:/.efs/$pl32_2->{name}/$dev_release},
            qq{Created directory $ldc_name:/.efs/$pl32_2->{name}/$dev_install},
            qq{Created symlink $ldc_name:/.efs/$pl32_2->{name}/$dev_install/exec -> /efs/$dev_install/.exec/$pl32_2->{name}},
        ],
        [
            [qw( none backwards)],
            qq{Created directory $ldc_name:/.efs/$pl64_1->{name}/$dev_release},
            qq{Created directory $ldc_name:/.efs/$pl64_1->{name}/$dev_install},
            qq{Created symlink $ldc_name:/.efs/$pl64_1->{name}/$dev_install/exec -> /efs/$dev_install/.exec/$pl64_1->{name}},
        ],
        [
            [qw( none )],
            qq{Created directory $ldc_name:/.efs/$pl64_2->{name}/$dev_release},
            qq{Created directory $ldc_name:/.efs/$pl64_2->{name}/$dev_install},
            qq{Created symlink $ldc_name:/.efs/$pl64_2->{name}/$dev_install/exec -> /efs/$dev_install/.exec/$pl64_2->{name}},
        ],
        [
            qq{Created symlink $ldc_name:/efs/$dev_release/install/exec -> /.efs/$dev_install/exec},
            qq{Created install /efs/$dev_install/.exec/$pl32_1->{name}},
        ],
    );

    my @output = ();

    foreach my $data ( @data ) {

        my @rows = @{ $data };

        if ( ref $rows[0] eq q{ARRAY} ) {
            my $config = shift @rows;
            if ( grep { $_ eq $release } @{ $config } ) {
                push @output, @rows;
            }
        } else {
            push @output, @rows;
        }

    }

    return @output;

}
